# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

"""Utilities for JSON to support PipeDAG objects

PipeDAG objects get serialized as JSON objects with a special `__pipedag_type__`
key to identify them when decoding. The associated value encodes the
type of the object.
"""

import datetime as dt
import importlib
import json
from enum import Enum
from functools import cache
from pathlib import Path
from typing import get_args, get_origin

from pydiverse.common.util.computation_tracing import fully_qualified_name
from pydiverse.common.util.import_ import load_object

TYPE_KEY = "__pipedag_type__"


class Type(str, Enum):
    # Pipedag types
    TABLE = "table"
    RAW_SQL = "raw_sql"
    BLOB = "blob"
    STAGE = "stage"
    PIPEDAG_CONFIG = "pipedag_config"
    CONFIG_CONTEXT = "config_context"

    # Data classes can be reconstructed
    DATA_CLASS = "data_class"

    # Types / Type Annotations can be reconstructed
    TYPE = "type"
    GENERIC_ALIAS = "generic_alias"

    # Other types we want to support
    PATHLIB_PATH = "pathlib:path"
    DT_DATE = "dt:date"
    DT_DATETIME = "dt:datetime"

    def __str__(self):
        return self.value


def json_default(o):
    """Encode `Table`, `RawSql`, 'Stage', and `Blob` objects"""
    from pydiverse.pipedag import ConfigContext, Stage
    from pydiverse.pipedag.container import Blob, RawSql, Table
    from pydiverse.pipedag.core.config import PipedagConfig

    if isinstance(o, Table):
        kwargs = {}
        if o.assumed_dependencies is not None:
            kwargs["assumed_dependencies"] = o.assumed_dependencies  # [json_default(t) for t in o.assumed_dependencies]
        return {
            TYPE_KEY: Type.TABLE,
            "stage": o.stage.name if o.stage is not None else None,
            "name": o.name,
            "primary_key": o.primary_key,
            "indexes": o.indexes,
            "cache_key": o.cache_key,
            "materialization_details": o.materialization_details,
            "external_schema": o.external_schema,
            "shared_lock_allowed": o.shared_lock_allowed,
            "annotation": o.annotation,
            **kwargs,
        }
    if isinstance(o, RawSql):
        kwargs = {}
        if o.assumed_dependencies is not None:
            kwargs["assumed_dependencies"] = o.assumed_dependencies
        return {
            TYPE_KEY: Type.RAW_SQL,
            "stage": o.stage.name if o.stage is not None else None,
            "name": o.name,
            "cache_key": o.cache_key,
            "table_names": o.table_names,
            **kwargs,
        }
    if isinstance(o, Blob):
        return {
            TYPE_KEY: Type.BLOB,
            "stage": o.stage.name if o.stage is not None else None,
            "name": o.name,
            "cache_key": o.cache_key,
        }
    if isinstance(o, Stage):
        return {
            TYPE_KEY: Type.STAGE,
            "name": o.name,
        }
    if isinstance(o, PipedagConfig):
        return {
            TYPE_KEY: Type.PIPEDAG_CONFIG,
            "config_dict": o.config_dict,
        }
    if isinstance(o, ConfigContext):
        return {
            TYPE_KEY: Type.CONFIG_CONTEXT,
            "config_dict": o._config_dict,
            "pipedag_name": o.pipedag_name,
            "flow_name": o.flow_name,
            "instance_name": o.instance_name,
        }
    if isinstance(o, Path):
        return {
            TYPE_KEY: Type.PATHLIB_PATH,
            "path": str(o),
        }
    if isinstance(o, dt.datetime):
        return {
            TYPE_KEY: Type.DT_DATETIME,
            "datetime": o.isoformat(),
        }
    if isinstance(o, dt.date):
        return {
            TYPE_KEY: Type.DT_DATE,
            "date": o.isoformat(),
        }
    if get_origin(o) is not None:
        # must be GenericAlias
        # somehow isinstance(o, GenericAlias) did not work reliably
        return {
            TYPE_KEY: Type.GENERIC_ALIAS,
            "origin": get_origin(o),
            "args": get_args(o),
        }
    if isinstance(o, type):
        return {
            TYPE_KEY: Type.TYPE,
            "module": o.__module__,
            "qualname": o.__qualname__,
        }
    try:
        # provide better error message in case of pdt.Table
        import pydiverse.transform as pdt

        if isinstance(o, pdt.Table):
            raise TypeError(
                "pydiverse.transform.Table is not supposed to be JSON serialized. "
                "It should be a pipedag Table instead."
                "Consider adding pydiverse.transform.Table to your auto_table setting "
                "in the PipedagConfig."
            )
    except ImportError:
        pass
    if hasattr(o, "__dataclass_fields__") and not isinstance(o, type):
        return {
            TYPE_KEY: Type.DATA_CLASS,
            "config_dict": {
                "class": fully_qualified_name(o),
                "args": o.__dict__,
            },
        }
    try:
        # provide better error message in case of pdt.Table
        import polars as pl

        if isinstance(o, pl.DataFrame | pl.LazyFrame):
            raise TypeError(
                "Polars Tables are not supposed to be JSON serialized. "
                "It should be a pipedag Table instead."
                "Consider adding polars.DataFrame and polars.LazyFrame to your "
                "auto_table setting in the PipedagConfig."
            )
    except ImportError:
        pass
    import pandas as pd

    if isinstance(o, pd.DataFrame):
        raise TypeError(
            "Pandas Tables are not supposed to be JSON serialized. "
            "It should be a pipedag Table instead."
            "Consider adding pandas.DataFrame to your "
            "auto_table setting in the PipedagConfig."
        )

    raise TypeError(f"Object of type {type(o)} is not JSON serializable")


def json_object_hook(d: dict):
    """Decode json with `Table` and `Blob` objects"""
    from pydiverse.pipedag import ConfigContext
    from pydiverse.pipedag.container import Blob, RawSql, Table
    from pydiverse.pipedag.context import RunContext
    from pydiverse.pipedag.core.config import PipedagConfig

    if TYPE_KEY not in d:
        return d

    type_ = Type(d[TYPE_KEY])

    def get_stage(name: str | None):
        if name is None:
            return None
        stages = RunContext.get().flow.stages
        return stages[name]

    if type_ == Type.TABLE:
        tbl = Table(
            name=d["name"],
            primary_key=d["primary_key"],
            indexes=d["indexes"],
            materialization_details=d.get("materialization_details"),
            annotation=d.get("annotation"),
        )
        tbl.stage = get_stage(d["stage"])
        tbl.cache_key = d["cache_key"]
        tbl.external_schema = d.get("external_schema")
        tbl.shared_lock_allowed = d.get("shared_lock_allowed", True)
        tbl.assumed_dependencies = d.get("assumed_dependencies")
        return tbl
    if type_ == Type.RAW_SQL:
        raw_sql = RawSql(name=d["name"])
        raw_sql.stage = get_stage(d["stage"])
        raw_sql.cache_key = d["cache_key"]
        raw_sql.table_names = d["table_names"]
        raw_sql.assumed_dependencies = d.get("assumed_dependencies")
        return raw_sql
    if type_ == Type.BLOB:
        blob = Blob(name=d["name"])
        blob.stage = get_stage(d["stage"])
        blob.cache_key = d["cache_key"]
        return blob
    if type_ == Type.STAGE:
        return get_stage(d["name"])
    if type_ == Type.PIPEDAG_CONFIG:
        return PipedagConfig(d["config_dict"])
    if type_ == Type.CONFIG_CONTEXT:
        return ConfigContext.new(d["config_dict"], d["pipedag_name"], d["flow_name"], d["instance_name"])
    if type_ == Type.PATHLIB_PATH:
        return Path(d["path"])
    if type_ == Type.DT_DATE:
        return dt.date.fromisoformat(d["date"])
    if type_ == Type.DT_DATETIME:
        return dt.datetime.fromisoformat(d["datetime"])
    if type_ == Type.DATA_CLASS:
        return load_object(d["config_dict"])
    if type_ == Type.TYPE:
        module = d["module"]
        qualname = d["qualname"]
        return getattr(importlib.import_module(module), qualname)
    if type_ == Type.GENERIC_ALIAS:
        origin = d["origin"]
        args = tuple(d["args"])
        return origin[args]

    raise ValueError(f"Invalid value for '{TYPE_KEY}' key: {type_}")


@cache
class PipedagJSONEncoder(json.JSONEncoder):
    def __init__(self):
        super().__init__(
            ensure_ascii=False,
            allow_nan=False,
            separators=(",", ":"),
            sort_keys=True,
            default=json_default,
        )


@cache
class PipedagJSONDecoder(json.JSONDecoder):
    def __init__(self):
        super().__init__(object_hook=json_object_hook)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import copy
import dataclasses
import importlib
import re
import types
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Any

import numpy as np
import pandas as pd
import sqlalchemy as sa
import sqlalchemy.dialects.mssql
from pandas.core.dtypes.base import ExtensionDtype

from pydiverse.common import Datetime, Dtype
from pydiverse.pipedag.backend.table.sql.ddl import (
    AddClusteredColumnstoreIndex,
    ChangeColumnTypes,
    CreateAlias,
    _mssql_update_definition,
)
from pydiverse.pipedag.backend.table.sql.hooks import (
    IbisTableHook,
    PandasTableHook,
    PolarsTableHook,
)
from pydiverse.pipedag.backend.table.sql.reflection import PipedagMSSqlReflection
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.container import RawSql, Schema, Table
from pydiverse.pipedag.materialize.details import (
    BaseMaterializationDetails,
    resolve_materialization_details_label,
)

try:
    from sqlalchemy import URL, Connection, Engine
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import URL, Engine
    from sqlalchemy.engine.base import Connection


@dataclass(frozen=True)
class MSSqlMaterializationDetails(BaseMaterializationDetails):
    """
    Materialization details specific to Microsoft SQL Server.

    :param columnstore: Whether to create tables as columnstores
        by defining a clustered columnstore index on them. This can
        improve performance.
        See the `Microsoft documentation <https://learn.microsoft.com/en-us/sql/relational-databases/indexes/columnstore-indexes-overview>`_.
    """

    def __post_init__(self):
        assert isinstance(self.columnstore, bool)

    columnstore: bool = False


class MSSqlTableStore(SQLTableStore):
    """
    SQLTableStore that supports `Microsoft SQL Server`_.

    Requires ODBC driver for Microsoft SQL Server to be installed.
    https://learn.microsoft.com/en-us/sql/connect/odbc/download-odbc-driver-for-sql-server
    For default arguments, both msodbcsql and mssql-tools need to be installed.
    Debug installation issues with `odbcinst -j`. On Linux, conda-forge pyodbc package
    expects ini files in different location than Microsoft installs it (symlink helps).

    Supports the :py:class:`MSSqlMaterializationDetails\
    <pydiverse.pipedag.backend.table.sql.dialects.mssql.MSSqlMaterializationDetails>`
    materialization details.

    In addition to the arguments of
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`,
    it also takes the following arguments:

    :param disable_pytsql:
        For mssql, a package called `pytsql <https://pypi.org/project/pytsql/>`_ is
        used for executing RawSql scripts. It has the advantage that it allows for
        some kind of SQL based print statements. However, it may fail for some
        statements. For those cases, you can set ``disable_pytsql: true`` to use another
        logic for splitting up Raw SQL scripts and handing that over to sqlalchemy.
        This is actually quite a complex process for mssql.
        Sorry for any inconveniences. We will try to make it work for most tsql code
        that should be integrated in pipedag pipelines. However, the ultimate goal
        is to split up monolithic blocks of dynamic sql statements into defined
        transformations with dynamic aspects written in python.

    :param disable_arrow_odbc:
        By default we try to use arrow-odbc to read data from Microsoft SQL Server.
        However, pyarrow has problems with long string columns.
        If you want to disable this, set this parameter to ``True``.

    :param disable_bulk_insert:
        By default we try to use bulk insert to write data to Microsoft SQL Server.
        Unfortunately, the bcp command line tool and bulk insert libraries like
        bcpandas have trouble with corner cases like linebreaks or other special
        characters in string columns.
        If you want to disable this, set this parameter to ``True``.
        If installed, bulk insert is done with mssqlkit (not open source). Otherwise,
        bcpandas is used.

    :param pytsql_isolate_top_level_statements:
        This parameter is handed over to `pytsql.executes`_ and causes the script to
        be split in top level statements that are sent to sqlalchemy separately.
        The tricky part here is that some magic is done to make DECLARE statements
        reach across, but it is not guaranteed to be identical to scripts executed
        by a SQL UI.

    :param max_query_print_length:
        Maximum number of characters in a SQL query that are printed when
        logging the query before the log is truncated. Defaults to 5000 characters.

    .. _Microsoft SQL Server:
        https://www.microsoft.com/en/sql-server/
    .. _pytsql.executes:
        https://pytsql.readthedocs.io/en/latest/api/pytsql.html#pytsql.executes
    """

    _dialect_name = "mssql"

    def __init__(
        self,
        *args,
        disable_pytsql: bool = False,
        disable_arrow_odbc: bool = False,
        disable_bulk_insert: bool = False,
        pytsql_isolate_top_level_statements: bool = True,
        max_query_print_length: int = 500000,
        **kwargs,
    ):
        self.disable_pytsql = disable_pytsql
        self.disable_arrow_odbc = disable_arrow_odbc
        self.disable_bulk_insert = disable_bulk_insert
        self.pytsql_isolate_top_level_statements = pytsql_isolate_top_level_statements
        self.max_query_print_length = max_query_print_length

        super().__init__(*args, **kwargs)

    def _init_database(self):
        self._init_database_with_database("master", {"isolation_level": "AUTOCOMMIT"})

    def dialect_requests_empty_creation(self, table: Table, is_sql: bool) -> bool:
        _ = is_sql
        return (
            table.nullable is not None
            or table.non_nullable is not None
            or (table.primary_key is not None and len(table.primary_key) > 0)
            or (table.indexes is not None and len(table.indexes) > 0)
            or self.get_create_columnstore_index(table)
        )

    def get_forced_nullability_columns(
        self, table: Table, table_cols: Iterable[str], report_nullable_cols=False
    ) -> tuple[list[str], list[str]]:
        # mssql dialect has literals as non-nullable types by default, so we also need
        # the list of nullable columns as well
        return self._process_table_nullable_parameters(table, table_cols)

    def add_indexes_and_set_nullable(
        self,
        table: Table,
        schema: Schema,
        *,
        on_empty_table: bool | None = None,
        table_cols: Iterable[str] | None = None,
    ):
        if on_empty_table is None or on_empty_table:
            # Set non-nullable and primary key on empty table
            key_columns = (
                [table.primary_key]
                if isinstance(table.primary_key, str)
                else table.primary_key
                if table.primary_key is not None
                else []
            )
            # reflect sql types because mssql cannot just change nullable flag
            inspector = sa.inspect(self.engine)
            columns = inspector.get_columns(table.name, schema=schema.get())
            table_cols = [d["name"] for d in columns]
            types = {d["name"]: d["type"] for d in columns}
            nullable_cols, non_nullable_cols = self.get_forced_nullability_columns(table, table_cols)
            non_nullable_cols = [col for col in non_nullable_cols if col not in key_columns]
            sql_types = [types[col] for col in nullable_cols]
            if len(nullable_cols) > 0:
                self.execute(
                    ChangeColumnTypes(
                        table.name,
                        schema,
                        nullable_cols,
                        sql_types,
                        nullable=True,
                    )
                )
            sql_types = [types[col] for col in non_nullable_cols]
            if len(non_nullable_cols) > 0:
                self.execute(
                    ChangeColumnTypes(
                        table.name,
                        schema,
                        non_nullable_cols,
                        sql_types,
                        nullable=False,
                    )
                )
            if len(key_columns) > 0:
                sql_types = [types[col] for col in key_columns]
                # impose some varchar(max) limit to allow use in primary key / index
                # TODO: consider making cap_varchar_max a config option
                self.execute(
                    ChangeColumnTypes(
                        table.name,
                        schema,
                        key_columns,
                        sql_types,
                        nullable=False,
                        cap_varchar_max=1024,
                    )
                )
                index_columns = set()  # type: set[str]
                if table.indexes is not None:
                    for index in table.indexes:
                        index_columns |= set(index)
                index_columns_list = list(index_columns)
                sql_types = self.reflect_sql_types(index_columns_list, table.name, schema)
                index_str_max_columns = [
                    col
                    for _type, col in zip(sql_types, index_columns_list)
                    if isinstance(_type, sa.String) and _type.length is None
                ]
                if len(index_str_max_columns) > 0:
                    # impose some varchar(max) limit to allow use in primary key / index
                    self.execute(
                        ChangeColumnTypes(
                            table.name,
                            schema,
                            index_str_max_columns,
                            sql_types,
                            cap_varchar_max=1024,
                        )
                    )
            if self.get_create_columnstore_index(table):
                self.execute(AddClusteredColumnstoreIndex(table.name, schema))
        if on_empty_table is None or not on_empty_table:
            self.add_table_primary_key(table, schema)
            self.add_table_indexes(table, schema)

    def execute_raw_sql(self, raw_sql: RawSql):
        if self.disable_pytsql:
            self.__execute_raw_sql_fallback(raw_sql)
        else:
            self.__execute_raw_sql_pytsql(raw_sql)

    def __execute_raw_sql_fallback(self, raw_sql: RawSql):
        """Alternative to using pytsql for splitting SQL statement.

        Known Problems:
        - DECLARE statements will not be available across GO
        """
        last_use_statement = None
        for statement in re.split(r"\bGO\b", raw_sql.sql, flags=re.IGNORECASE):
            statement = statement.strip()
            if statement == "":
                # allow GO at end of script
                continue

            if re.match(r"\bUSE\b", statement):
                last_use_statement = statement

            with self.engine_connect() as conn:
                if last_use_statement is not None:
                    self.execute(last_use_statement, conn=conn)
                self.execute(statement, conn=conn)
                if last_use_statement is not None:
                    # ensure database connection is reset to original database
                    self.execute(f"USE {self.engine_url.database}", conn=conn)

    def __execute_raw_sql_pytsql(self, raw_sql: RawSql):
        """Use pytsql for executing T-SQL scripts containing many GO statements."""
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import pytsql

        sql_string = raw_sql.sql
        if self.print_sql:
            print_query_string = self.format_sql_string(sql_string)
            if len(print_query_string) >= self.max_query_print_length:
                print_query_string = print_query_string[: self.max_query_print_length] + " [...]"
            self.logger.info("Executing sql", query=print_query_string)
        # ensure database connection is reset to original database
        sql_string += f"\nUSE {self.engine_url.database}"

        pytsql.executes(
            sql_string,
            self.engine,
            isolate_top_level_statements=self.pytsql_isolate_top_level_statements,
        )

    def _get_all_objects_in_schema(self, schema: Schema) -> dict[str, Any]:
        with self.engine_connect() as conn:
            # tables (U), views (V), synonyms (SN), procedures (P) and functions (FN)
            query = """
                SELECT obj.name, obj.type
                FROM sys.objects AS obj
                LEFT JOIN sys.schemas AS schem ON schem.schema_id = obj.schema_id
                WHERE obj.type IN ('U', 'V', 'SN', 'P', 'FN')
                  AND schem.name = :schema
                  AND obj.is_ms_shipped = 0
            """
            result = conn.execute(sa.text(query), {"schema": schema.get()}).all()
        return {name: type_.strip() for name, type_ in result}

    def _copy_object_to_transaction(
        self,
        name: str,
        metadata: Any,
        src_schema: Schema,
        dest_schema: Schema,
        conn: Connection,
    ):
        type_: str = metadata

        if type_ == "SN":
            alias_name, alias_schema = self.resolve_alias(name, src_schema.get())
            self.execute(
                CreateAlias(
                    alias_name,
                    Schema(alias_schema, "", ""),
                    name,
                    dest_schema,
                )
            )
            return

        # For views, procedures and functions we need to inspect the definition
        # of the object and replace all references to the src schema with
        # references to the dest schema.
        assert type_ in ("V", "P", "FN")
        definition = _mssql_update_definition(conn, name, src_schema, dest_schema)
        self.execute(definition, conn=conn)

    def resolve_alias(self, table: Table, stage_name: str):
        # The base implementation already takes care of converting Table objects
        # based on ExternalTableReference objects to string table name and schema.
        # For normal Table objects, it needs the stage schema name.
        table_name, schema = super().resolve_alias(table, stage_name)
        return PipedagMSSqlReflection.resolve_alias(self.engine, table_name, schema)

    def _set_materialization_details(self, materialization_details: dict[str, dict[str | list[str]]] | None) -> None:
        self.materialization_details = MSSqlMaterializationDetails.create_materialization_details_dict(
            materialization_details,
            self.strict_materialization_details,
            self.default_materialization_details,
            self.logger,
        )

    def get_create_columnstore_index(self, table: Table) -> bool:
        return MSSqlMaterializationDetails.get_attribute_from_dict(
            self.materialization_details,
            resolve_materialization_details_label(table),
            self.default_materialization_details,
            "columnstore",
            self.strict_materialization_details,
            self.logger,
        )


try:
    import polars as pl
except ImportError:
    pl = importlib.import_module("polars")
    pl.DataType = None
    pl.DataFrame = None
    pl.LazyFrame = None


class DataframeMsSQLTableHook:
    @classmethod
    def dialect_has_adbc_driver(cls):
        # MSSQL does not support ADBC drivers, so far.
        return False

    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) supports MSSQL
        return True

    @classmethod
    def _get_dialect_dtypes(
        cls, dtypes: dict[str, Dtype], table: Table[pd.DataFrame]
    ) -> dict[str, sa.types.TypeEngine]:
        sql_dtypes = ({name: dtype.to_sql() for name, dtype in dtypes.items()}) | (
            {name: sa.dialects.mssql.DATETIME2() for name, dtype in dtypes.items() if dtype == Datetime()}
        )
        if table.type_map:
            sql_dtypes.update(table.type_map)
        return sql_dtypes

    @classmethod
    def upload_table_bulk_insert(
        cls,
        table: Table[pd.DataFrame | pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        df = table.obj  # type: pd.DataFrame | pl.DataFrame
        schema_name = schema.get()

        mssqlkit = False
        try:
            import quantcore.mssqlkit as mss

            mssqlkit = True
        except ImportError:
            store.logger.debug(
                "mssqlkit not installed, falling back to bcpandas. This is expected since mssqlkit is not open source."
            )

        if mssqlkit:
            mss.table.bulk_upload(
                engine=store.engine,
                table_name=f"{schema_name}.{table.name}",
                data=df,
            )
        else:
            # use bcpandas
            from bcpandas import SqlCreds, to_sql

            if len(df) == 0:
                # bcpands won't write an empty table
                # (this method is inherited from PandasTableHook/PolarsTableHook)
                cls._dialect_create_empty_table(store, table, schema, dtypes)
            else:
                creds = SqlCreds.from_engine(store.engine)
                url = store.engine_url.render_as_string()
                if "&" in url:
                    creds.odbc_kwargs = {
                        x.split("=")[0]: x.split("=")[1] for x in url.split("&")[1:] if len(x.split("=")) == 2
                    }
                if isinstance(df, pl.DataFrame):
                    # convert polars DataFrame to pandas DataFrame
                    df = df.to_pandas()
                # attention: the `(lambda col: lambda...)(copy.copy(col))` part looks odd
                # but is needed to ensure loop iterations create lambdas working on
                # different columns
                df_out = df.assign(
                    **{
                        col: (lambda col: lambda df: df[col].map({True: 1, False: 0}).astype(pd.Int8Dtype()))(
                            copy.copy(col)
                        )
                        for col, dtype in df.dtypes[(df.dtypes == "bool[pyarrow]") | (df.dtypes == "bool")].items()
                    }
                ).assign(
                    **{
                        col: (
                            lambda col: lambda df: df[col]
                            .str.replace("\n", "\\n", regex=False)
                            .str.replace("\t", "\\t", regex=False)
                            .str.replace("\r", "\\r", regex=False)
                        )(copy.copy(col))
                        for col, dtype in df.dtypes[
                            (df.dtypes == "object") | (df.dtypes == "string") | (df.dtypes == "string[pyarrow]")
                        ].items()
                    }
                )
                to_sql(
                    df_out,
                    table.name,
                    creds,
                    schema=schema_name,
                    index=False,
                    if_exists="append" if early else "fail",
                    batch_size=min(len(df), 100_000),
                    dtype=dtypes,
                    delimiter="\t",
                    quotechar="\1",
                )

    @classmethod
    def download_table_arrow_odbc(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, pl.DataFrame] | None = None,
    ) -> pl.DataFrame:
        assert dtypes is None, (
            "Pyarrow reads SQL schema and loads the data in reasonable types."
            "Thus, manual dtype manipulation can only done via query or afterwards."
        )
        engine = store.engine

        odbc_string = ConnectionString.from_url(engine.url).odbc_string

        # we need to avoid VARCHAR(MAX) columns since arrow-odbc does not
        # handle them well
        # dtypes = {}  # TODO: allow user to provide custom dtypes
        # column_types = reflect_pyodbc_column_types(query, odbc_string)

        df = pl.concat(
            pl.read_database(
                query=query,
                connection=odbc_string,
                iter_batches=True,
                batch_size=65536,
                # schema_overrides=dtypes,
                execute_options={"map_schema": map_pyarrow_schema_for_polars},
            ),
            rechunk=True,
        )

        return df

    @classmethod
    def _adjust_cols_retrieve(cls, store: MSSqlTableStore, cols: dict, dtypes: dict) -> tuple[dict | None, dict]:
        assert isinstance(store, MSSqlTableStore)
        arrow_odbc = not store.disable_arrow_odbc
        max_string_length = 256
        if arrow_odbc:
            # arrow-odbc does not handle VARCHAR(MAX) columns well, so we need to cut
            # strings
            if any(isinstance(col.type, sa.String) for col in cols.values()):
                return {
                    name: sa.cast(
                        sa.func.substring(col, 1, max_string_length).label(name),
                        sa.String(max_string_length),
                    )
                    if isinstance(col.type, sa.String)
                    else col
                    for name, col in cols.items()
                }, dtypes
        return None, dtypes


@MSSqlTableStore.register_table(pd)
class PandasTableHook(DataframeMsSQLTableHook, PandasTableHook):
    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
    ) -> pd.DataFrame:
        assert isinstance(store, MSSqlTableStore)
        arrow_odbc = not store.disable_arrow_odbc

        if arrow_odbc:
            try:
                pl_df = cls.download_table_arrow_odbc(query, store, dtypes=None)
                df = pl_df.to_pandas()
                df = cls._fix_dtypes(df, dtypes)
            except Exception as e:  # noqa
                store.logger.warning("Failed to download table using arrow-odbc, falling back to sqlalchemy/pandas.")

        return super().download_table(query, store, dtypes)

    @classmethod
    def upload_table(
        cls,
        table: Table[pd.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        if not store.disable_bulk_insert:
            try:
                return cls.upload_table_bulk_insert(table, schema, dtypes, store, early)
            except Exception as e:  # noqa
                store.logger.exception("Failed to upload table using bulk insert, falling back to pandas.")
        # TODO: consider using arrow-odbc for uploading
        super().upload_table(table, schema, dtypes, store, early)


def reflect_pyodbc_column_types(query: str, odbc_string: str):
    """Reflect the column types of a query using pyodbc."""
    import pyodbc

    with pyodbc.connect(odbc_string) as conn:
        with conn.cursor() as cursor:
            cursor.execute(f"SELECT TOP 0 * FROM ({query}) as A")
            column_types = [
                {
                    "name": d[0],  # column name
                    "sql_type": d[1],  # ODBC SQL type code, e.g. -5 for BIGINT
                    "precision": d[4],  # column_size
                    "scale": d[5],  # decimal_digits
                    "nullable": d[6],  # 1 = nullable, 0 = NOT NULL
                }
                for d in cursor.description
            ]
    return column_types


@MSSqlTableStore.register_table(pl.DataFrame)
class PolarsTableHook(DataframeMsSQLTableHook, PolarsTableHook):
    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, pl.DataFrame] | None = None,
    ) -> pl.DataFrame:
        assert dtypes is None, (
            "Pyarrow reads SQL schema and loads the data in reasonable types."
            "Thus, manual dtype manipulation can only be done via query or afterwards."
        )
        assert isinstance(store, MSSqlTableStore)
        arrow_odbc = not store.disable_arrow_odbc

        if arrow_odbc:
            try:
                return cls.download_table_arrow_odbc(query, store, dtypes)
            except Exception as e:  # noqa
                store.logger.warning(
                    "Failed to download table using arrow-odbc, falling back to "
                    "sqlalchemy/polars which currently uses pandas in the back."
                )

        return super().download_table(query, store, dtypes)

    @classmethod
    def upload_table(
        cls,
        table: Table[pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        """
        Provide hook that allows to override the default
        upload of polars tables to the tablestore.
        """
        if not store.disable_bulk_insert:
            try:
                return cls.upload_table_bulk_insert(table, schema, dtypes, store, early)
            except:  # noqa
                store.logger.warning("Failed to upload table using bulk insert, falling back to polars.write_database.")
        super().upload_table(table, schema, dtypes, store, early)


try:
    import ibis
except ImportError:
    ibis = None


@MSSqlTableStore.register_table(ibis)
class IbisTableHook(IbisTableHook):
    @classmethod
    def _conn(cls, store: MSSqlTableStore):
        url = store.engine_url
        return ibis.mssql.connect(
            host=url.host,
            user=url.username,
            password=url.password,
            port=url.port,
            database=url.database,
        )


try:
    import pyarrow as pa
except ImportError:
    pa = types.ModuleType("pyarrow")
    pa.Schema = None


def map_pyarrow_schema_for_polars(schema: pa.Schema) -> pa.Schema:
    """This is necessary since, by default, arrow-odbc loads all DATETIME2 columns with
    nanosecond precision before potentially casting them to microsecond precision (as
    specified in the polars schema), which results in overflows.

    Hence, we have to tell arrow-odbc to load the fields only in microsecond precision.
    """
    assert pa.schema is not None, "please install pyarrow"
    mapped_fields = [
        (field.with_type(pa.timestamp("us")) if field.type == pa.timestamp("ns") else field) for field in list(schema)
    ]
    return pa.schema(mapped_fields, schema.metadata)


DEFAULT_PORT = 1433
MSSQL_DRIVER_REGEX = re.compile(r"ODBC Driver ([0-9]+) for SQL Server")
ODBC_CONNECTION_FIELD_KEYMAP = {
    "server": "host",
    "pwd": "password",
    "uid": "username",
    "trustservercertificate": "insecure",
}
TRUTH_STR_VALUES = ("y", "yes", "t", "true", "True", "on", "1", "1.0")
FALSE_STR_VALUES = ("n", "no", "f", "false", "False", "off", "0", "0.0")

TRUTH_VALUES = (*TRUTH_STR_VALUES, True, 1.0)
FALSE_VALUES = (*FALSE_STR_VALUES, False, 0.0)

NA_VALUES: list[Any] = ["nan", "NaN", "NaT", "NULL", "null", float("nan")]


def _latest_mssql_driver() -> str:
    import pyodbc

    # Get all drivers
    drivers = pyodbc.drivers()

    # Filter by SQL drivers and get their versions
    mssql_driver_versions = []
    for driver in drivers:
        match = MSSQL_DRIVER_REGEX.match(driver)
        if match:
            mssql_driver_versions.append(int(match.groups()[0]))
    if len(mssql_driver_versions) == 0:
        raise ValueError("No ODBC driver for SQL Server found.")

    # Return driver name of latest version
    latest_version = max(mssql_driver_versions)
    return f"ODBC Driver {latest_version} for SQL Server"


@dataclass
class ConnectionString:
    """Utility class for handling parameters of a database connection string.

    You typically initialize this connection string manually by passing values for at
    least ``host`` and ``database``. Afterwards, you will want to either...

    - create a :class:`~sqlalchemy.Engine` by using :meth:`sqlalchemy_url` or
    - create a connection with e.g. :mod:`pyodbc` by using ``str(self)``

    Note:
        You often have to either disable encryption or allow insecure connections
        within corporate networks.
    """

    #: The host of the database instance. Might be a hostname or an IP address.
    host: str
    #: The name of the database to connect to in the database instance.
    database: str = "tempdb"
    #: The port to connect to.
    port: int | None = DEFAULT_PORT
    #: The name of the ODBC driver to use for connecting to the database.
    driver: str = field(default_factory=_latest_mssql_driver)

    #: The username for connecting to the database (if not using Windows Auth).
    username: str | None = None
    #: The password of the user to authenticate with (if not using Windows Auth).
    password: str | None = None

    #: Whether to encrypt communication with the database in exchange for performance.
    encrypt: bool = True
    #: Whether to skip server certificate validation for TLS-encrypted communication.
    insecure: bool = False

    @classmethod
    def from_engine(cls, engine: Engine) -> "ConnectionString":
        return ConnectionString.from_url(engine.url)

    @classmethod
    def from_url(cls, url: URL) -> "ConnectionString":
        """Parse a ConnectionString from an url.

        Args:
            url: URL to parse; both sqlalchemy and odbc url formats are valid

        Returns:
            ConnectionString parsed from url
        """
        # url is a sqlalchemy url
        query = {k.lower(): v for k, v in url.query.items()}

        kwargs = {}
        if query.get("driver") is not None:
            kwargs["driver"] = query["driver"]
        if query.get("encrypt") is not None:
            kwargs["encrypt"] = (
                query["encrypt"].lower() not in FALSE_STR_VALUES  # type: ignore
            )
        if query.get("trustservercertificate") is not None:
            kwargs["insecure"] = (
                query["trustservercertificate"].lower()  # type: ignore
                in TRUTH_STR_VALUES
            )

        return ConnectionString(
            host=url.host or "localhost",
            database=url.database or "tempdb",
            port=url.port,
            username=url.username,
            password=url.password if url.username is not None else None,
            **kwargs,  # type: ignore
        )

    @classmethod
    def _from_odbc_string(cls, odbc_string: str):
        """Parse a ConnectionString from the fields of an odbc dictionary string.

        Args:
            odbc_string: odbc connection dictionary string

        Returns:
            ConnectionString parsed from the odbc connection arguments
        """
        kwargs: dict[str, Any] = {}
        for entry in odbc_string.split(";"):
            key_val = entry.split("=")
            assert len(key_val) == 2
            adjusted_key = ODBC_CONNECTION_FIELD_KEYMAP.get(key_val[0].lower(), key_val[0].lower())
            if adjusted_key == "trusted_connection":
                continue  # we can determine this key from username presence
            kwargs[adjusted_key] = key_val[1]

        if "driver" in kwargs:
            kwargs["driver"] = kwargs["driver"].replace("{", "").replace("}", "")
        kwargs["encrypt"] = kwargs.get("encrypt", "").lower() not in FALSE_STR_VALUES
        kwargs["insecure"] = kwargs.get("insecure", "").lower() in TRUTH_STR_VALUES

        return ConnectionString(
            **kwargs,
        )

    def with_database(self, /, database: str) -> "ConnectionString":
        """Obtain a copy of the connection string reference a different database.

        The returned connection string connects to the same database instance as the
        original connection string and uses the same connection parameters.

        Args:
            database: The name of the database that the connection string should
                reference.

        Returns:
            The new connection string.
        """
        return dataclasses.replace(self, database=database)

    @property
    def driver_version(self) -> int:
        """The version of the ODBC driver referenced by the connection string."""
        match = MSSQL_DRIVER_REGEX.match(self.driver)
        if match is None:
            raise ValueError(f"Cannot determine driver version from '{self.driver}'.")
        return int(match.groups()[0])

    @property
    def odbc_string(self) -> str:
        """The ODBC connection string to use for :mod:`pyodbc` and :mod:`arrow-odbc`."""
        params = {
            "Driver": "{" + self.driver + "}",
            "Server": (f"{self.host},{self.port}" if self.port else self.host),
            "Database": self.database,
            "Encrypt": "yes" if self.encrypt else "no",
            "TrustServerCertificate": "yes" if self.insecure else "no",
        }
        if self.username is None:
            params["Trusted_Connection"] = "yes"
        else:
            params["UID"] = self.username
            if self.password is not None:
                params["PWD"] = self.password
        return ";".join(f"{k}={v}" for k, v in params.items())

    def fine_grained_odbc_string(self, multiple_active_result_sets: bool = False) -> str:
        """Construct an ODBC connection string with more more fine-grained control."""
        return self.odbc_string + f";MARS_Connection={'yes' if multiple_active_result_sets else 'no'}"

    @property
    def sqlalchemy_url(self) -> URL:
        """The connection string as a URL for SQLAlchemy.

        This URL should be used when calling :meth:`sqlalchemy.create_engine`.

        Note:
            This method intentionally does NOT create a URL with the format
            ``mssql+pyodbc:///?odbc_connect={odbc_args}`` since this does not allow for
            accessing the ``database`` property of an engine URL in SQLAlchemy.
        """
        query = {
            "driver": self.driver,
            "Encrypt": "yes" if self.encrypt else "no",
            "TrustServerCertificate": "yes" if self.insecure else "no",
        }
        if self.username is None:
            query["Trusted_Authentication"] = "yes"
        return URL.create(
            drivername="mssql+pyodbc",
            username=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
            query=query,
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import dataframely as dy
import pandas as pd
import polars as pl
import sqlalchemy as sa

import pydiverse.colspec as cs
import pydiverse.transform as pdt
from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import AUTO_VERSION, Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.transform.extended import left_join


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        ),
        name="lazy_1",
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.Alias, input2: sa.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"])


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return tbl1.merge(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_polars(tbl1: pl.LazyFrame, tbl2: pl.LazyFrame):
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_pdt(tbl1: pdt.Table, tbl2: pdt.Table):
    return tbl1 >> left_join(tbl2, tbl1.x == tbl2.x) >> pdt.alias("eager_task_pdt_out")


@materialize(lazy=True, input_type=pdt.SqlAlchemy)
def lazy_task_pdt(tbl1: pdt.Table, tbl2: pdt.Table):
    # pydiverse.transform syntax works both on Polars and SQL/duckdb backends.
    return tbl1 >> left_join(tbl2, tbl1.x == tbl2.x) >> pdt.alias("lazy_task_pdt_out")


class Tbl1Schema(dy.Schema):
    x = dy.Int32()
    y = dy.Int32()


class OutputSchema(Tbl1Schema):
    a = dy.Int64()


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_dataframely(tbl1: dy.LazyFrame[Tbl1Schema], tbl2: pl.LazyFrame) -> OutputSchema:
    # Pipedag automatically calls Tbl1Schema.cast() and OutputSchema.validate().
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_dataframely_pdt(tbl1: Tbl1Schema, tbl2: pdt.Table) -> OutputSchema:
    # Pipedag automatically calls Tbl1Schema.cast() and OutputSchema.validate().
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("eager_task_dataframely_pdt_out")
    )


class Tbl1ColSpec(cs.ColSpec):
    x = cs.Int32()
    y = cs.Int32()


class OutputColSpec(Tbl1ColSpec):
    a = cs.Int64()


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_colspec(tbl1: Tbl1ColSpec, tbl2: pl.LazyFrame) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast_polars() and OutputColSpec.validate_polars()
    # which call dataframely in the background. ColSpec classes can also be used for SQL validations.
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_colspec_pdt(tbl1: Tbl1ColSpec, tbl2: pdt.Table) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast_polars() and OutputColSpec.validate_polars()
    # which call dataframely in the background. ColSpec classes can also be used for SQL validations.
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("eager_task_colspec_pdt_out")
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_colspec(tbl1: Tbl1ColSpec, tbl2: sa.Alias) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast() and OutputColSpec.validate()
    return Table(
        sa.select(tbl1, tbl2.c.a).select_from(tbl1.outerjoin(tbl2, tbl1.c.x == tbl2.c.x)), name="lazy_task_colspec"
    )


@materialize(lazy=True, input_type=pdt.SqlAlchemy)
def lazy_task_colspec_pdt(tbl1: Tbl1ColSpec, tbl2: pdt.Table) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast() and OutputColSpec.validate()
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("lazy_task_colspec_pdt_out")
    )


def main():
    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            a, b = eager_inputs()

        with Stage("stage_2"):
            lazy_2 = lazy_task_2(lazy_1, b)
            lazy_3 = lazy_task_3(lazy_2)
            # Do the same operation with various syntax alternatives.
            # This is also a demonstration how pipedag interacts with \
            # dataframely, pydiverse.colspec, and pydiverse.transform.
            out = eager_task(lazy_1, b)
            out2 = eager_task_polars(lazy_1, b)
            out3 = eager_task_pdt(lazy_1, b)
            out4 = lazy_task_pdt(lazy_1, b)
            out5 = eager_task_dataframely(lazy_1, b)
            out6 = eager_task_dataframely_pdt(lazy_1, b)
            out7 = eager_task_colspec(lazy_1, b)
            out8 = eager_task_colspec_pdt(lazy_1, b)
            out9 = lazy_task_colspec(lazy_1, b)
            out10 = lazy_task_colspec_pdt(lazy_1, b)

        with Stage("stage_3"):
            lazy_4 = lazy_task_4(lazy_2)
        _ = lazy_3, lazy_4, out, out2, out3, out4, out5, out6, out7, out8, out9, out10  # unused terminal output tables

    # Run flow
    result = f.run()
    assert result.successful

    # Run in a different way for testing
    with StageLockContext():
        result = f.run()
        assert result.successful
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish

    # see https://pydiversepipedag.readthedocs.io/en/latest/table_backends.html#installing-mssql-odbc-driver-for-macos-and-linux
    # for how to install the ODBC driver for MSSQL on macOS and Linux.

    # You can run this pipeline with:
    # ```shell
    # pixi run python run_pipeline.py
    # ```

    # don't forget to run docker-compose in separate shell to launch minio container:
    # ```shell
    # pixi run docker-compose up
    # ```

    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# this code is inspired by https://github.com/dmgass/baseline/ but uses separate JSON
# files and supports parameterized tests
import json
import os.path
import traceback
from pathlib import Path
from typing import Any

import structlog


class BaselineStore:
    def __init__(self, *args, **kwargs):
        self.caller_frame = traceback.extract_stack(limit=2)[-2]
        self.compare_file = Path(self.caller_frame.filename).with_suffix(".json")
        if kwargs:
            self.parameters = kwargs.copy()
            if args:
                self.parameters.update({f"_arg{i}": args[i] for i in range(len(args))})
        else:
            self.parameters = list(args)
        self.last_cmp = None
        self.logger = structlog.get_logger(self.__class__.__name__, parameters=self.parameters)

    def __eq__(self, other):
        self.last_cmp = None
        exists = False
        if self.compare_file.is_file():
            exists = True
            with open(self.compare_file) as f:
                baseline = json.load(f)
            updated_baseline = baseline.copy()
        else:
            updated_baseline = {}
        if self.parameters:
            cmp = self.swap(key_path=self.parameters, value=other, inout=updated_baseline)
        else:
            cmp = updated_baseline if updated_baseline else None
            updated_baseline = other
        updated_file = self.compare_file.with_suffix(".updated.json")
        if exists and updated_file.is_file() and os.path.getmtime(updated_file) > os.path.getmtime(self.compare_file):
            # Update the update_file instead of self.compare_file since it is newer.
            # Like this, parameterized tests can accumulate their joint update.
            with open(updated_file) as f:
                updated_baseline2 = json.load(f)
            self.swap(key_path=self.parameters, value=other, inout=updated_baseline2)
            updated_baseline = updated_baseline2
        with open(updated_file, "w") as f:
            json.dump(updated_baseline, f, indent=4)
        if not exists:
            # write a base file that can be easily diffed against updated file
            with open(self.compare_file, "w") as f:
                if isinstance(updated_baseline, dict):
                    json.dump({}, f, indent=4)
                else:
                    json.dump(None, f, indent=4)
        self.last_cmp = cmp
        if not (cmp == other):
            self.logger.info(f"Comparison is false: BaselineStore({self.parameters}):\n   '{cmp}'\n!= '{other}'")
        return cmp == other

    def __repr__(self):
        if self.last_cmp:
            return f"BaselineStore({self.parameters})={self.last_cmp}"
        else:
            return f"BaselineStore({self.parameters})"

    @staticmethod
    def swap(key_path: list | dict[str, Any], value, *, inout):
        key_path = key_path.copy()
        mutable_dict = inout
        if not key_path:
            return value
        if isinstance(key_path, dict):
            try_key = list(mutable_dict.keys())[0] if mutable_dict else None
            if try_key in key_path:
                key_value = str(key_path.pop(try_key))
                if key_value in mutable_dict[try_key]:
                    if key_path:
                        return BaselineStore.swap(key_path, value, inout=mutable_dict[try_key][key_value])
                    return BaselineStore._swap_value(key_value, value, inout=mutable_dict[try_key])
                return BaselineStore._add_dict_key(try_key, key_value, key_path, value, inout=mutable_dict)
            else:
                key, key_value = key_path.popitem()
                key, key_value = str(key), str(key_value)
                return BaselineStore._add_dict_key(key, key_value, key_path, value, inout=mutable_dict)
        elif isinstance(key_path, list):
            key = str(key_path.pop(0))
            if key in mutable_dict:
                if key_path:
                    return BaselineStore.swap(key_path, value, inout=mutable_dict[key])
                else:
                    return BaselineStore._swap_value(key, value, inout=mutable_dict)
            else:
                if key_path:
                    mutable_dict[key] = {}
                    return BaselineStore.swap(key_path, value, inout=mutable_dict[key])
                else:
                    mutable_dict[key] = value
                    return None
        else:
            raise TypeError(f"Fatal error: BaselineStore messed up self.parameters to include type: {type(key_path)}")

    @staticmethod
    def _swap_value(key, value, *, inout):
        mutable_dict = inout
        save = mutable_dict[key]
        mutable_dict[key] = value
        return save

    @staticmethod
    def _add_dict_key(key, key_value, key_path, value, *, inout):
        mutable_dict = inout
        if key_path:
            mutable_dict[key][key_value] = {}
            return BaselineStore.swap(key_path, value, mutable_dict[key][key_value])
        mutable_dict[key][key_value] = value
        return None  # no match found



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import pytest

from pydiverse.pipedag import Flow, Stage, Table, materialize

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances
from tests.util.tasks_library import assert_table_equal

# unfortunately, pydiverse.transform currently does not support ibm_db2
pytestmark = [pytest.mark.pdtransform, with_instances(tuple(set(DATABASE_INSTANCES) - {"ibm_db2"}))]

try:
    import pydiverse.transform as pdt

    _ = pdt

    try:
        from pydiverse.transform.core.verbs import mutate
        from pydiverse.transform.eager import PandasTableImpl
        from pydiverse.transform.lazy import SQLTableImpl

        # ensures a "used" state for the import, preventing black from deleting it
        _ = PandasTableImpl

        test_list = [SQLTableImpl, PandasTableImpl]
    except ImportError:
        try:
            from pydiverse.transform.extended import Pandas, Polars, SqlAlchemy, mutate

            test_list = [SqlAlchemy, Polars, Pandas]
        except ImportError:
            raise NotImplementedError("pydiverse.transform 0.2.0 - 0.2.2 isn't supported") from None
except ImportError:
    test_list = []


@pytest.mark.parametrize(
    "impl_type",
    test_list,
)
def test_table_store(impl_type: type):
    def cache_fn(*args, **kwargs):
        return impl_type.__name__

    @materialize()
    def in_table():
        return Table(
            pd.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                }
            )
        )

    @materialize()
    def expected_out_table():
        return Table(
            pd.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                    "x": [1, 1, 1, 1],
                    "y": [2, 2, 2, 2],
                }
            )
        )

    @materialize(input_type=impl_type, cache=cache_fn)
    def noop(x):
        return Table(x)

    @materialize(lazy=True, input_type=impl_type, cache=cache_fn)
    def noop_lazy(x):
        return Table(x)

    @materialize(input_type=impl_type, cache=cache_fn)
    def add_column(x):
        return Table(x >> mutate(x=1))

    @materialize(lazy=True, input_type=impl_type, cache=cache_fn)
    def add_column_lazy(x):
        return Table(x >> mutate(y=2))

    with Flow() as f:
        with Stage("pdtransform"):
            table = in_table()
            table = noop(table)
            table = noop_lazy(table)
            table = add_column(table)
            table = add_column_lazy(table)

            expected = expected_out_table()
            _ = assert_table_equal(table, expected, check_dtype=False)

    assert f.run().successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag.backend.table.sql.dialects.duckdb_parquet import (
    ParquetTableStore,
)

__all__ = [
    "ParquetTableStore",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import os
import shutil
from typing import Any

import sqlalchemy as sa
from upath import UPath

import pydiverse.pipedag.backend.table.sql.hooks as sql_hooks
from pydiverse.pipedag import ConfigContext, Schema, Stage, Table
from pydiverse.pipedag.backend.table.sql.ddl import (
    CopySelectTo,
    CreateViewAsSelect,
    DropView,
)
from pydiverse.pipedag.backend.table.sql.dialects.duckdb import DuckDBTableStore
from pydiverse.pipedag.backend.table.sql.sql import DISABLE_DIALECT_REGISTRATION
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.materialize.table_hook_base import (
    AutoVersionSupport,
    CanMatResult,
    CanRetResult,
    TableHook,
)
from pydiverse.pipedag.util import normalize_name
from pydiverse.pipedag.util.path import is_file_uri

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import duckdb
except ImportError:
    duckdb = None

try:
    from sqlalchemy import Select, TextClause
    from sqlalchemy import Text as SqlText
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.sql.expression import TextClause
    from sqlalchemy.sql.selectable import Select

    SqlText = TextClause  # this is what sa.text() returns


class ParquetTableStore(DuckDBTableStore):
    """Table store that materializes tables as parquet files.

    Additionally, a duckdb file is created which links to parquet files.

    Uses schema swapping for transactions:
    Creates a schema for each stage and a temporary schema for each
    transaction. If all tasks inside a stage succeed, swaps the schemas by
    renaming them.

    .. rubric:: Supported Tables

    The `SQLTableStore` can materialize (that is, store task output)
    and dematerialize (that is, retrieve task input) the following Table types:

    .. list-table::
       :widths: 30, 30, 30
       :header-rows: 1
       :stub-columns: 1

       * - Framework
         - Materialization
         - Dematerialization

       * - SQLAlchemy
         - | :py:class:`sa.sql.expression.Selectable
                        <sqlalchemy.sql.expression.Selectable>`
           | :py:class:`sa.sql.expression.TextClause
                        <sqlalchemy.sql.expression.TextClause>`
         - :py:class:`sa.Table <sqlalchemy.schema.Table>`

       * - Pandas
         - :py:class:`pd.DataFrame <pandas.DataFrame>`
         - :py:class:`pd.DataFrame <pandas.DataFrame>`

       * - Polars
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`
         - | :external+pl:doc:`pl.DataFrame <reference/dataframe/index>`
           | :external+pl:doc:`pl.LazyFrame <reference/lazyframe/index>`

       * - tidypolars
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`
         - :py:class:`tp.Tibble <tidypolars.tibble.Tibble>`

       * - Ibis
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`
         - :py:class:`ibis.api.Table <ibis.expr.types.relations.Table>`

       * - pydiverse.transform
         - ``pdt.Table``
         - | ``pdt.eager.PandasTableImpl``
           | ``pdt.lazy.SQLTableImpl``

       * - pydiverse.pipedag table reference
         - :py:class:`~.ExternalTableReference` (no materialization)
         - Can be read with all dematerialization methods above

    :param url:
        The :external+sa:ref:`SQLAlchemy engine url <database_urls>`
        used to connect to the database.

        This URL may contain placeholders like ``{name}`` or ``{instance_id}``
        (additional ones can be defined in the ``url_attrs_file``) or
        environment variables like ``{$USER}`` which get substituted with their
        respective values.

        Attention: passwords including special characters like ``@`` or ``:`` need
        to be URL encoded.

    :param url_attrs_file:
        Filename of a yaml file which is read shortly before rendering the final
        engine URL and which is used to replace custom placeholders in ``url``.

        Just like ``url``, this value may also contain placeholders and environment
        variables which get substituted.

    :param create_database_if_not_exists:
        If the engine url references a database name that doesn't yet exists,
        then setting this value to ``True`` tells pipedag to create the database
        before trying to open a connection to it.

    :param schema_prefix:
        A prefix that gets placed in front of all schema names created by pipedag.

    :param schema_suffix:
        A suffix that gets placed behind of all schema names created by pipedag.

    :param avoid_drop_create_schema:
        If ``True``, no ``CREATE SCHEMA`` or ``DROP SCHEMA`` statements get issued.
        This is mostly relevant for databases that support automatic schema
        creation like IBM DB2.

    :param print_materialize:
        If ``True``, all tables that get materialized get logged.

    :param print_sql:
        If ``True``, all executed SQL statements get logged.

    :param no_db_locking:
        Speed up database by telling it we will not rely on it's locking mechanisms.
        Currently not implemented.

    :param strict_materialization_details:
        If ``True``: raise an exception if
            - the argument ``materialization_details`` is given even though the
              table store does not support it.
            - a table references a ``materialization_details`` tag that is not defined
              in the config.
        If ``False``: Log an error instead of raising an exception

    :param materialization_details:
        A dictionary with each entry describing a tag for materialization details of
        the table store. See subclasses of :py:class:`BaseMaterializationDetails
         <pydiverse.pipedag.materialize.details.BaseMaterializationDetails>`
        for details.
    :param default_materialization_details:
        The materialization_details that will be used if materialization_details
        is not specified on table level. If not set, the ``__any__`` tag (if specified)
        will be used.
    :param max_concurrent_copy_operations:
        In case of a partially cache-valid stage, we need to copy tables from the
        cache schema to the new transaction schema. This parameter specifies the
        maximum number of workers we use for concurrently copying tables.
    :param sqlalchemy_pool_size:
        The number of connections to keep open inside the connection pool.
        It is recommended to choose a larger number than
        ``max_concurrent_copy_operations`` to avoid running into pool_timeout.
    :param sqlalchemy_pool_timeout:
        The number of seconds to wait before giving up on getting a connection from
        the pool. This may be relevant in case the connection pool is saturated
        with concurrent operations each working with one or more database connections.
    :param parquet_base_bath:
        This is an fsspec compatible path to either a directory or an S3 bucket key
        prefix. Examples are '/tmp/pipedag/parquet/' or 's3://pipedag-test-bucket/table_store/'.
        `instance_id` will automatically be appended to parquet_base_bath.

    """

    _dialect_name = DISABLE_DIALECT_REGISTRATION

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        config = config.copy()
        instance_id = normalize_name(ConfigContext.get().instance_id)
        config["parquet_base_path"] = UPath(config["parquet_base_path"]) / instance_id
        return super()._init_conf_(config)

    def __init__(
        self,
        engine_url: str,
        parquet_base_path: UPath,
        *,
        create_database_if_not_exists: bool = False,
        schema_prefix: str = "",
        schema_suffix: str = "",
        avoid_drop_create_schema: bool = False,
        print_materialize: bool = False,
        print_sql: bool = False,
        no_db_locking: bool = True,
        strict_materialization_details: bool = True,
        materialization_details: dict[str, dict[str | list[str]]] | None = None,
        default_materialization_details: str | None = None,
        max_concurrent_copy_operations: int = 5,
        sqlalchemy_pool_size: int = 12,
        sqlalchemy_pool_timeout: int = 300,
    ):
        super().__init__(
            engine_url,
            create_database_if_not_exists=create_database_if_not_exists,
            schema_prefix=schema_prefix,
            schema_suffix=schema_suffix,
            avoid_drop_create_schema=avoid_drop_create_schema,
            print_materialize=print_materialize,
            print_sql=print_sql,
            no_db_locking=no_db_locking,
            strict_materialization_details=strict_materialization_details,
            materialization_details=materialization_details,
            default_materialization_details=default_materialization_details,
            max_concurrent_copy_operations=max_concurrent_copy_operations,
            sqlalchemy_pool_size=sqlalchemy_pool_size,
            sqlalchemy_pool_timeout=sqlalchemy_pool_timeout,
        )
        self.parquet_base_path = parquet_base_path

        # ## state

        # whether parquet files of stage are stored in __odd or __even
        # can typically be found in metadata, but firstly we cache this
        # and secondly we modify it for the current stage before commit
        self.parquet_schema_paths: dict[str, UPath] = {}
        # tables may get alias to other transaction slot in case of cache
        # valid tasks
        self.parquet_table_paths: dict[tuple[str, str], UPath] = {}
        # remember tables that need to be copied if stage is not unchanged
        self.parquet_deferred_copy: list[Table] = []

        self.logger.info(
            "Initialized Parquet Table Store",
            parquet_base_path=self.parquet_base_path,
            engine_url=self.engine_url.render_as_string(hide_password=True),
            schema_prefix=self.schema_prefix,
            schema_suffix=self.schema_suffix,
        )

    def _create_engine(self):
        engine = super()._create_engine(
            connect_args={
                # configure s3 for minio (see docker-compose.yaml)
                "config": {
                    "s3_endpoint": "localhost:9000",
                    "s3_url_style": "path",
                    "s3_use_ssl": "false",
                    # "s3_region": "us-east-1",
                    "s3_access_key_id": "minioadmin",
                    "s3_secret_access_key": "minioadmin",
                },
            }
        )

        def execute(con, statement):
            """
            Execute a SQL statement and return the result.
            """
            if sa.__version__ >= "2.0.0":
                return con.execute(sa.text(statement))
            else:
                return con.execute(statement)

        with engine.connect() as con:
            # Enable HTTPFS extension
            execute(con, "INSTALL httpfs;")
            execute(con, "LOAD httpfs;")

            # MinIO credentials (default unless you've changed them)
            execute(con, "SET s3_access_key_id='minioadmin';")
            execute(con, "SET s3_secret_access_key='minioadmin';")

            # MinIO-specific settings
            execute(con, "SET s3_use_ssl=false;")
            execute(con, "SET s3_endpoint='localhost:9000';")  # MinIO server URL
            execute(con, "SET s3_url_style='path';")  # Important: MinIO uses path-style URLs

            if sa.__version__ >= "2.0.0":
                con.commit()

        return engine

    def init_stage(self, stage: Stage):
        # fetch existing tables from database before schema is deleted there
        old_transaction_name = self._get_read_view_transaction_name(stage.name)
        new_transaction_name = self._get_read_view_original_transaction_name(stage, old_transaction_name)
        with self.engine_connect() as conn:
            existing_tables = self.execute(
                f"FROM duckdb_views() SELECT view_name WHERE "
                f"schema_name='{new_transaction_name}' "
                f"and sql like '%%FROM read_parquet(%%'",
                conn=conn,
            ).fetchall()
        # this creates transaction schema and modifies stage.current_name
        super().init_stage(stage)
        # update cache: this is important since before this stage commits,
        # self._get_read_view_transaction_name() will yield the wrong name
        new_schema = self.get_schema(stage.current_name)
        self.parquet_schema_paths[new_schema.name] = self.get_parquet_path(new_schema)
        # The stage.name should still point to opposite transaction schema
        old_schema = self.get_schema(self._get_read_view_original_transaction_name(stage))
        self.parquet_schema_paths[self.get_schema(stage.name).name] = self.get_parquet_path(old_schema)
        # TODO: for nested stages this might be per stage
        self.parquet_deferred_copy.clear()
        path = self.get_stage_path(stage)
        if is_file_uri(path):
            os.makedirs(path, exist_ok=True)
        for row in existing_tables:
            table = row[0]
            file_path = path / (table + ".parquet")
            try:
                self.logger.info(
                    "Cleaning up parquet file in transaction schema",
                    file_path=file_path,
                )
                file_path.unlink()
            except FileNotFoundError:
                self.logger.error(
                    "Could not remove parquet file while deleting corresponding view in transaction schema",
                    file=file_path,
                    view_name=table,
                    stage=stage,
                )

    def _copy_table(self, table: Table, from_schema: Schema, from_name: str):
        if table.name != from_name:
            # we ignore the _copy_table which wants to write __copy tables
            return
        _ = from_schema  # not used because this only points to READ VIEW
        dest_schema = self.get_schema(table.stage.current_name)
        original_transaction_name = self._get_read_view_original_transaction_name(table.stage)
        original_schema = self.get_schema(original_transaction_name)
        src_file_path = self.get_parquet_path(original_schema) / (from_name + ".parquet")
        dest_file_path = self.get_parquet_path(dest_schema) / src_file_path.name
        self.logger.info(
            "Copying table between transactions",
            src_file_path=src_file_path,
            dest_file_path=dest_file_path,
        )
        if is_file_uri(src_file_path):
            shutil.copy(src_file_path, dest_file_path)
        else:
            src_file_path.fs.copy(src_file_path.as_uri(), dest_file_path.as_uri(), on_error="raise")
        # create view in duckdb database file
        self.execute(CreateViewAsSelect(table.name, dest_schema, self._read_parquet_query(dest_file_path)))

    @staticmethod
    def _read_parquet_query(file_path):
        # attention: we filter for %%FROM read_parquet(%% in init_stage
        return sa.text(f"FROM read_parquet('{file_path}')")

    def _swap_alias_with_table_copy(self, table: Table, table_copy: Table):
        # There is no __copy table which can be renamed for this table store.
        # Alias will be replaced in commit_stage()
        pass

    def _deferred_copy_table(
        self,
        table: Table,
        from_schema: Schema,
        from_name: str,
    ):
        # this will just create an alias in the duckdb database
        super()._deferred_copy_table(table, from_schema, from_name)
        # keep symlink to original parquet file until stage commits
        original_transaction_name = self._get_read_view_original_transaction_name(table.stage)
        original_schema = self.get_schema(original_transaction_name)
        schema = self.get_schema(table.stage.current_name)
        self.parquet_table_paths[(schema.name, table.name)] = self.get_parquet_path(original_schema) / (
            table.name + ".parquet"
        )
        self.parquet_deferred_copy.append(table)

    def commit_stage(self, stage: Stage):
        schema = self.get_schema(stage.current_name)
        stage_transaction_path = self.parquet_schema_paths[schema.name]
        # deferred copy of parquet files if stage is not 100% cache valid
        if RunContext.get().has_stage_changed(stage):
            for table in self.parquet_deferred_copy:
                # remove views which were placed as aliases
                self.execute(DropView(table.name, schema))
                src_file_path = self.parquet_table_paths[(schema.name, table.name)]
                dest_file_path = stage_transaction_path / src_file_path.name
                self.logger.info(
                    "Copying table between transactions",
                    src_file_path=src_file_path,
                    dest_file_path=dest_file_path,
                )
                if is_file_uri(src_file_path):
                    shutil.copy(src_file_path, dest_file_path)
                else:
                    src_file_path.fs.copy(src_file_path.as_uri(), dest_file_path.as_uri(), on_error="raise")
                # replace view in duckdb database file (which was previously just
                # alias to main schema)
                self.execute(CreateViewAsSelect(table.name, schema, self._read_parquet_query(dest_file_path)))
            # switch committed transaction path to the new transaction path
            self.parquet_schema_paths[self.get_schema(stage.name).name] = stage_transaction_path
        super().commit_stage(stage)
        # clear table individual parquet paths since this is not needed after commit
        self.parquet_table_paths = {}

    def _committed_unchanged(self, stage):
        super()._committed_unchanged(stage)
        # transaction schema is discarded, so we don't need to copy tables there
        self.parquet_deferred_copy.clear()

    def get_parquet_path(self, schema: Schema):
        return self.parquet_base_path / schema.get()

    def get_parquet_schema_path(self, schema: Schema) -> UPath:
        # Parquet files are stored in transaction schema while stage.current_name
        # changes to final schema. We resolve a level of indirection in memory.
        if schema.name in self.parquet_schema_paths:
            return self.parquet_schema_paths[schema.name]
        # now we assume schema.name == stage.name (any stage with
        # current_name != name should be in self.parquet_schema_paths)
        transaction_name = self._get_read_view_transaction_name(schema.name)
        if transaction_name == "":
            transaction_name = schema.name
        path = self.get_parquet_path(self.get_schema(transaction_name))
        self.parquet_schema_paths[schema.name] = path
        return path

    def get_stage_path(self, stage: Stage) -> UPath:
        store = ConfigContext.get().store.table_store
        return self.get_parquet_schema_path(store.get_schema(stage.current_name))

    def get_table_path(
        self,
        table: Table,
        file_extension: str = ".parquet",
        lookup_schema: bool = False,
    ) -> UPath:
        store = ConfigContext.get().store.table_store
        schema_name = table.stage.current_name
        return self.get_table_schema_path(table.name, store.get_schema(schema_name), file_extension)

    def get_table_schema_path(self, table_name: str, schema: Schema, file_extension: str = ".parquet") -> UPath:
        # Parquet files might be stored in other transaction schema in case of cache
        # validity. We resolve a level of indirection in memory.
        if (schema.name, table_name) in self.parquet_table_paths:
            return self.parquet_table_paths[(schema.name, table_name)]
        return self.get_parquet_schema_path(schema) / (table_name + file_extension)

    def delete_table_from_transaction(self, table: Table, *, schema: Schema | None = None):
        if schema is None:
            schema = self.get_schema(table.stage.transaction_name)
        file_path = self.get_table_schema_path(table.name, schema)
        file_path.unlink(missing_ok=True)
        self.execute(
            DropView(
                table.name,
                self.get_schema(table.stage.transaction_name),
                if_exists=True,
            )
        )

    def add_primary_key(
        self,
        table_name: str,
        schema: Schema,
        key_columns: list[str],
        *,
        name: str | None = None,
    ):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def add_index(
        self,
        table_name: str,
        schema: Schema,
        index_columns: list[str],
        name: str | None = None,
    ):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def copy_indexes(self, src_table: str, src_schema: Schema, dest_table: str, dest_schema: Schema):
        # ParquetTableStore does not support indexes. They are ignored.
        pass

    def rename_table(self, table: Table, to_name: str, schema: Schema):
        _ = schema  # ignore given schema
        schema = self.get_schema(table.stage.transaction_name)
        from_path = self.get_table_schema_path(table.name, schema)
        to_path = self.get_table_schema_path(to_name, schema)
        if is_file_uri(from_path):
            os.rename(from_path, to_path)
        else:
            from_path.fs.copy(from_path.as_uri(), to_path.as_uri(), on_error="raise")
            from_path.unlink()

        # drop view
        self.execute(
            DropView(
                table.name,
                schema,
            )
        )
        # create view again
        self.execute(CreateViewAsSelect(to_name, schema, self._read_parquet_query(to_path)))

    def write_subquery(
        self,
        query,
        to_name,
        neighbor_table: Table,
        *,
        unlogged: bool = False,
        suffix: str | None = None,
    ):
        """Write a query to a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        _ = unlogged, suffix  # materialization details not used by ParquetTableStore
        schema = self.get_schema(neighbor_table.stage.transaction_name)
        file_path = self.get_table_schema_path(to_name, schema)
        self.execute(
            [
                CopySelectTo(
                    file_path,
                    "PARQUET",
                    query,
                ),
                CreateViewAsSelect(to_name, schema, self._read_parquet_query(file_path)),
            ]
        )
        return schema

    def drop_subquery_table(
        self,
        drop_name: str,
        schema: Schema | str,
        neighbor_table: Table,
        if_exists: bool = False,
        cascade: bool = False,
    ):
        """Drop a table in same schema as neighbor_table.

        This is mainly for overriding by derived table stores like ParquetTableStore.
        """
        _ = schema  # ignore given schema; use neighbor_table schema instead
        _ = cascade
        schema = self.get_schema(neighbor_table.stage.transaction_name)
        drop_path = self.get_table_schema_path(drop_name, schema)
        try:
            drop_path.unlink()
        except FileNotFoundError as e:
            if not if_exists:
                raise e
        # drop view
        self.execute(
            DropView(
                drop_name,
                schema,
            )
        )


@ParquetTableStore.register_table(pd)
class PandasTableHook(TableHook[ParquetTableStore]):
    auto_version_support = AutoVersionSupport.TRACE

    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def materialize(
        cls,
        store: ParquetTableStore,
        table: Table[pd.DataFrame],
        stage_name: str,
        without_config_context: bool = False,
    ):
        file_path = store.get_table_path(table)
        schema = store.get_schema(table.stage.current_name)

        if store.print_materialize:
            store.logger.info(
                f"Writing pandas table '{schema.get()}.{table.name}'",
                file_path=file_path,
                table_obj=table.obj,
            )

        df = table.obj
        df.to_parquet(file_path)
        store.execute(CreateViewAsSelect(table.name, schema, store._read_parquet_query(file_path)))

    @classmethod
    def retrieve(
        cls,
        store: ParquetTableStore,
        table: Table,
        stage_name: str | None,
        as_type: type,
        limit: int | None = None,
    ):
        path = store.get_table_path(table, lookup_schema=True)
        import pyarrow.dataset as ds

        if limit is not None:
            df = ds.dataset(path).scanner().head(limit).to_pandas()
        else:
            df = pd.read_parquet(path)
        import pyarrow.parquet

        schema = pyarrow.parquet.read_schema(path)
        df = df.astype(
            {
                col: "datetime64[s]"
                for col, type_, dtype in zip(schema.names, schema.types, df.dtypes)
                if type_ == "date32" and dtype == object  # noqa: E721
            }
        )
        if isinstance(as_type, tuple) and len(as_type) == 2 and len(schema.names) > 0:
            if as_type[1] == "arrow" and not all(
                hasattr(dtype, "storage") and dtype.storage == "pyarrow" for dtype in df.dtypes
            ):
                store.logger.warning(
                    f"Ignoring storage specialization '{as_type[1]}' for {table.name} "
                    f"as ParquetTableStore reads pandas the same encoding as it writes"
                    f" it. This can be changed if need arises.",
                    as_type=as_type,
                    dtypes=df.dtypes,
                )
            if as_type[1] == "numpy" and any(
                hasattr(dtype, "storage") and dtype.storage == "pyarrow" for dtype in df.dtypes
            ):
                store.logger.warning(
                    f"Ignoring storage specialization '{as_type[1]}' for {table.name} "
                    f"as ParquetTableStore reads pandas the same encoding as it writes"
                    f" it. This can be changed if need arises.",
                    as_type=as_type,
                    dtypes=df.dtypes,
                )
        return df

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        return sql_hooks.PandasTableHook.auto_table(obj)

    @classmethod
    def get_computation_tracer(cls):
        return sql_hooks.PandasTableHook.ComputationTracer()


try:
    import polars as pl
except ImportError:
    pl = None


@ParquetTableStore.register_table(pl, duckdb)
class PolarsTableHook(sql_hooks.PolarsTableHook):
    def _execute_materialize_polars(table, store, stage_name):
        _ = stage_name
        file_path = store.get_table_path(table)
        schema = store.get_schema(table.stage.current_name)
        df = table.obj

        if store.print_materialize:
            store.logger.info(
                f"Writing polars table '{schema.get()}.{table.name}' to parquet",
                file_path=file_path,
            )
        df.write_parquet(file_path)
        store.execute(CreateViewAsSelect(table.name, schema, store._read_parquet_query(file_path)))

    @classmethod
    def _execute_query(
        cls,
        store: ParquetTableStore,
        table: Table,
        stage_name: str,
        as_type: type,
        limit: int | None = None,
    ) -> pl.DataFrame:
        _ = as_type
        file_path = store.get_table_path(table)
        df = pl.read_parquet(file_path, n_rows=limit)
        return df


@ParquetTableStore.register_table(pl)
class LazyPolarsTableHook(sql_hooks.LazyPolarsTableHook):
    pass


@ParquetTableStore.register_table()
class SQLAlchemyTableHook(sql_hooks.SQLAlchemyTableHook):
    @classmethod
    def _create_as_select_statements(
        cls,
        table_name: str,
        schema: Schema,
        query: Select | TextClause | SqlText,
        store: ParquetTableStore,
        suffix: str,
        unlogged: bool,
    ):
        file_path = store.get_table_schema_path(table_name, schema)
        return [
            CopySelectTo(
                file_path,
                "PARQUET",
                query,
            ),
            CreateViewAsSelect(table_name, schema, store._read_parquet_query(file_path)),
        ]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from .sesson import Session
from .util import parse_config


def pytest_addoption(parser):
    workers_help = "Set the max num of workers (aka processes) to start (int or 'auto' - one per core)"

    group = parser.getgroup("parallelize")
    group.addoption("--workers", dest="workers", help=workers_help)
    parser.addini("workers", workers_help)


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    workers = parse_config(config, "workers")
    if config.option.collectonly or not workers:
        return

    config.pluginmanager.register(Session(config), "parallelize-session")

    try:
        # Patch _jb_pytest_runner to support parallel execution of test
        # when using the PyCharm IDE
        from _jb_runner_tools import set_parallel_mode

        set_parallel_mode()
    except ImportError:
        pass


@pytest.hookimpl
def pytest_addhooks(pluginmanager):
    from . import hooks

    pluginmanager.add_hookspecs(hooks)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import atexit
import warnings
from typing import Any

from pydiverse.common.util import requires
from pydiverse.pipedag import ConfigContext, Stage
from pydiverse.pipedag.backend.lock.base import BaseLockManager, Lockable, LockState
from pydiverse.pipedag.errors import DisposedError, LockError
from pydiverse.pipedag.util import normalize_name

try:
    import kazoo
    from kazoo.client import KazooClient, KazooState
    from kazoo.recipe.lock import Lock as KazooLock
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    kazoo = None
    KazooClient = None


@requires(kazoo, ImportError("ZooKeeperLockManager requires 'kazoo' to be installed."))
class ZooKeeperLockManager(BaseLockManager):
    """Apache ZooKeeper based lock manager

    Uses Apache ZooKeeper to establish `fully distributed locks that are
    globally synchronous`_. The advantage of this approach is that we it is highly
    reliable and that in case our flow crashes, the acquired locks automatically
    get released (the locks are ephemeral).

    Config File
    -----------
    All arguments in the ``args`` section get passed as-is to the initializer
    of :py:class:`kazoo.client.KazooClient`. Some useful arguments include:

    :param hosts:
        Comma separated list of hosts to connect.

    .. _fully distributed locks that are globally synchronous:
        https://zookeeper.apache.org/doc/r3.1.2/recipes.html#sc_recipes_Locks
    """

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        client = KazooClient(**config)
        instance_id = normalize_name(ConfigContext.get().instance_id)
        base_path = f"/pipedag/locks/{instance_id}/"
        return cls(client, base_path)

    def __init__(self, client: KazooClient, base_path: str):
        super().__init__()

        self.client = client
        self.base_path = base_path
        if not self.client.connected:
            self.client.start()
            atexit.register(self.__atexit)
        self.client.add_listener(self._lock_listener)

        self.locks: dict[Lockable, KazooLock] = {}

    def __atexit(self):
        try:
            self.dispose()
        except DisposedError:
            pass

    def dispose(self):
        self.release_all()
        self.client.stop()
        self.client.close()
        self.lock_states.clear()
        self.locks.clear()
        super().dispose()

    @property
    def supports_stage_level_locking(self):
        return True

    def acquire(self, lockable: Lockable):
        lock = self.client.Lock(self.lock_path(lockable))
        self.logger.info(f"Locking '{lockable}'", base_path=self.base_path)
        if not lock.acquire():
            raise LockError(f"Failed to acquire lock '{lockable}'")
        self.locks[lockable] = lock
        self.set_lock_state(lockable, LockState.LOCKED)

    def release(self, lockable: Lockable):
        if lockable not in self.locks:
            raise LockError(f"No lock '{lockable}' found.")

        self.logger.info(f"Unlocking '{lockable}'", base_path=self.base_path)
        self.locks[lockable].release()
        del self.locks[lockable]
        self.set_lock_state(lockable, LockState.UNLOCKED)

    def lock_path(self, lock: Lockable):
        if isinstance(lock, Stage):
            return self.base_path + lock.name
        elif isinstance(lock, str):
            return self.base_path + lock
        else:
            raise NotImplementedError(f"Can't lock object of type '{type(lock).__name__}'")

    def _lock_listener(self, state):
        if state == KazooState.SUSPENDED:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.UNCERTAIN)
        elif state == KazooState.LOST:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.INVALID)
        elif state == KazooState.CONNECTED:
            for lock in self.locks.keys():
                self.set_lock_state(lock, LockState.LOCKED)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from upath import UPath


def is_file_uri(path: UPath):
    return path.as_uri().startswith("file://")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from multiprocessing import Queue

import pytest
import structlog
from _pytest.config import Config


def start_worker(worker_id: int, work_queue: Queue, msg_queue: Queue, args: list, option_dict: dict):
    option_dict["plugins"].append("no:terminal")
    config = Config.fromdictargs(option_dict, args)
    config.args = args

    from typing import TextIO

    class DontPrint(TextIO, ABC):
        def write(*_):
            pass

    # fix assert inspection code of pytest raised in threads
    # register dummy terminal reporter since it is needed by pytest even with
    # plugins:"no:terminal" option
    from _pytest.terminal import TerminalReporter

    terminal_reporter = TerminalReporter(config, DontPrint())
    config.pluginmanager.register(terminal_reporter, "terminalreporter")

    # Remove workers option to prevent triggering main plugin
    config.option.workers = None

    worker = Worker(config, worker_id, work_queue, msg_queue)
    config.pluginmanager.register(worker)
    config.hook.pytest_cmdline_main(config=config)


class Worker:
    def __init__(self, config, worker_id: int, work_queue: Queue, msg_queue: Queue):
        super().__init__()

        self.config = config
        self.worker_id = worker_id
        self.work_queue = work_queue
        self.msg_queue = msg_queue
        self.logger = structlog.get_logger(__name__, worker_id=worker_id)

        self.session_items = {}

    def send(self, msg, **kwargs):
        kwargs["worker_id"] = self.worker_id
        self.msg_queue.put((msg, kwargs))

    @pytest.hookimpl
    def pytest_sessionstart(self, session):
        self.send("sessionstart")

    @pytest.hookimpl
    def pytest_sessionfinish(self, session):
        self.send("sessionfinish")

    @pytest.hookimpl
    def pytest_runtest_logstart(self, nodeid, location):
        self.send("logstart", nodeid=nodeid, location=location)

    @pytest.hookimpl
    def pytest_runtest_logfinish(self, nodeid, location):
        self.send("logfinish", nodeid=nodeid, location=location)

    @pytest.hookimpl
    def pytest_runtest_logreport(self, report):
        data = self.config.hook.pytest_report_to_serializable(
            config=self.config,
            report=report,
        )
        self.send("logreport", report=data)

    @pytest.hookimpl
    def pytest_report_teststatus(self, report):
        """Overriding this hook is important with our custom terminal reporter.

        Otherwise, this hook will return None and multi-tests will abort after
        running first test. Unfortunately, this is silent and looks as everything
        is good. There were just hardly any tests run.
        """
        from _pytest.terminal import pytest_report_teststatus

        return pytest_report_teststatus(report)

    @pytest.hookimpl
    def pytest_runtestloop(self, session):
        self.session_items = {item.nodeid: item for item in session.items}
        should_terminate = False
        while not should_terminate:
            command, args = self.work_queue.get()
            should_terminate = self.process_one_item(session, command, args)
        return True

    def process_one_item(self, session, command, args):
        if command == "STOP":
            return True

        if command == "GROUP":
            group_name, node_ids = args
            items = [self.session_items[node_id] for node_id in node_ids]

            self.send("DEBUG_start_group", group_name=group_name)
            for i, item in enumerate(items):
                next_item = items[i + 1] if i + 1 < len(items) else None
                self.run_one_test(session, item, next_item)

        return False

    def run_one_test(self, session, item, next_item):
        self.send("DEBUG_start_test", nodeid=item.nodeid)
        item.ihook.pytest_runtest_protocol(item=item, nextitem=next_item)
        if session.shouldfail:
            raise session.Failed(session.shouldfail)
        if session.shouldstop:
            raise session.Interrupted(session.shouldstop)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# isort: skip_file
from .task import Task, UnboundTask
from .config import PipedagConfig
from .group_node import GroupNode, VisualizationStyle
from .stage import Stage
from .result import Result
from .flow import Flow, Subflow

__all__ = [
    "Flow",
    "Subflow",
    "PipedagConfig",
    "Result",
    "Stage",
    "GroupNode",
    "VisualizationStyle",
    "UnboundTask",
    "Task",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

try:
    import dataframely as dy
except ImportError:
    dy = None

try:
    import pydiverse.transform as pdt
except ImportError:
    pdt = None

try:
    import pydiverse.colspec as cs
except ImportError:
    cs = None

from example.run_pipeline import main as example_flow_main
from example.simple_pipeline import main as simple_pipeline_main
from example.visualization import main as visualization_main
from example_imperative.run_pipeline import main as example_imperative_flow_main
from example_interactive.run_tasks_interactively import main as example_interactive_main
from tests.fixtures.instances import with_instances

examples = [
    example_flow_main,
    simple_pipeline_main,
    visualization_main,
    example_imperative_flow_main,
    example_interactive_main,
]
s3_examples = []
mssql_examples = []
ibmdb_examples = []

if dy and cs and pdt:
    from example_ibm_db2.run_pipeline import main as example_ibmdb_flow_main
    from example_mssql.run_pipeline import main as example_mssql_flow_main
    from example_parquet_s3.run_pipeline import main as example_s3_main
    from example_postgres.run_pipeline import main as example_postgres_flow_main

    examples.append(example_postgres_flow_main)
    s3_examples.append(example_s3_main)
    mssql_examples.append(example_mssql_flow_main)
    ibmdb_examples.append(example_ibmdb_flow_main)


@pytest.mark.parametrize(
    "fn",
    examples,
)
@with_instances("postgres")
def test_examples(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    s3_examples,
)
@with_instances("parquet_s3_backend")
def test_examples_s3(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    mssql_examples,
)
@with_instances("mssql")
def test_examples_mssql(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()


@pytest.mark.parametrize(
    "fn",
    ibmdb_examples,
)
@with_instances("ibm_db2")
def test_examples_ibmdb(fn):
    """
    This test just runs the example pipeline that we provide in example/run_pipeline.py
    """

    fn()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import logging
import tempfile

import pandas as pd
import sqlalchemy as sa

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, Stage, Table, input_stage_versions, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.core.config import create_basic_pipedag_config


# The @input_stage_versions decorator offers all the options that @materialize has as
# well. `input_type=sa.Table` may be a good choice for tests on larger tables.
@input_stage_versions(input_type=pd.DataFrame)
def validate_stage1(tbls: dict[str, pd.DataFrame], other_tbls: dict[str, pd.DataFrame]):
    # Any tests can be done on tables of both versions. They can either fail and prevent
    # the schema swap, print some information about differences, or produce a table
    # with results of validation.
    assert tbls["task_1_out"]["x"][0] == 1
    assert tbls["task_2_out"]["a"].sum() == 3
    tables = {t.lower() for t in tbls.keys() if not t.endswith("__copy")}
    assert len(tables - {"task_1_out", "task_2_out", "dfa"}) == 1
    assert list(tables - {"task_1_out", "task_2_out", "dfa"})[0].startswith("dfb_")
    logger = logging.getLogger(f"{__name__}-validate_stage1")
    logger.info("Additional tables: %s", set(tbls.keys()) - set(other_tbls.keys()))
    logger.info(
        "Missing tables: %s",
        set(other_tbls.keys()) - set(tbls.keys()) - {"column_diffs"},
    )

    # Producing a table with differences of matching table names can be done a lot more
    # elaborate in a library. Here is just an idea to get started:
    def get_missing_columns(tbl: pd.DataFrame, other_tbl: pd.DataFrame):
        return set(other_tbl.columns) - set(tbl.columns)

    missing_columns = {
        tbl: get_missing_columns(tbls[tbl], other_tbls[tbl]) for tbl in set(tbls.keys()) & set(other_tbls.keys())
    }
    col_diff_dfs = [pd.DataFrame(dict(table=[], column=[], value=[]))]
    for tbl, columns in missing_columns.items():
        col_diff_df = pd.DataFrame(
            dict(
                table=tbl,
                column=list(columns),
                value="missing",
            )
        )
        col_diff_dfs.append(col_diff_df)
    col_diff_df = pd.concat(col_diff_dfs, ignore_index=True, axis="rows")
    return Table(col_diff_df, name="column_diffs")


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        ),
        name="task_1_out",
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.sql.expression.Alias, input2: sa.sql.expression.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out")


add_column = {"c": [1, 2, 3, 4]}


@materialize(nout=2)
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
            **add_column,
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


def main():
    logger = logging.getLogger(__name__)
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: If uncommented, stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            #   You need to install optional dependency 'pydot' for any visualization
            #   URL to appear.
            # kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as f:
                with Stage("stage_1"):
                    lazy_1 = lazy_task_1()
                    a, b = eager_inputs()
                    lazy_2 = lazy_task_2(lazy_1, b)
                    col_diff = validate_stage1()

                _ = a, lazy_2  # unused terminal output tables

            # Run Flow a bit different first
            result = f.run()
            assert result.successful

            # change something in table dfa
            global add_column
            add_column = {}

            # Run Flow and print diff result
            with StageLockContext():
                result = f.run()
                assert result.successful
                logger.info(
                    "Column differences:\n%s",
                    result.get(col_diff, as_type=pd.DataFrame),
                )


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import sqlalchemy as sa

from pydiverse.pipedag import Stage, materialize
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.context import ConfigContext, TaskContext


@materialize(input_type=sa.Table, lazy=True)
def sql_script(
    name: str,
    script_directory: Path,
    *,
    input_stage=None,
    depend=None,
):
    _ = depend  # only relevant for adding additional task dependency
    stage = TaskContext.get().task.stage

    script_path = script_directory / name
    sql = Path(script_path).read_text(encoding="utf-8")
    sql = raw_sql_bind_schema(sql, "out_", stage, transaction=True)
    sql = raw_sql_bind_schema(sql, "in_", input_stage)
    return RawSql(sql)


def raw_sql_bind_schema(sql, prefix: str, stage: Stage | RawSql | None, *, transaction=False):
    config = ConfigContext.get()
    store = config.store.table_store
    if stage is not None:
        stage_name = stage.transaction_name if transaction else stage.name
        schema_name = store.get_schema(stage_name).get()
        sql = sql.replace(f"{{{{{prefix}schema}}}}", schema_name)
    return sql



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Table, Task, materialize
from pydiverse.pipedag.core.task import TaskGetItem

try:
    import polars as pl
except ImportError:
    pl = None


def noop(x):
    # fail already at declare time
    assert isinstance(x, (Task, TaskGetItem))

    @materialize(input_type=pd.DataFrame, version="1.0")
    def _noop(x):
        # constant or collection of constants not supported in imperative version
        assert isinstance(x, pd.DataFrame)
        return Table(x).materialize()

    return _noop(x)


@materialize(input_type=sa.Table, version="1.0")
def noop_sql(x):
    return Table(x).materialize()


@materialize(input_type=sa.Table, lazy=True)
def noop_lazy(x):
    return Table(x).materialize()


@materialize(input_type=pl.DataFrame if pl else "<polars task without polars>")
def noop_polars(x):
    return Table(x).materialize()


@materialize(input_type=pl.LazyFrame if pl else "<polars task without polars>")
def noop_lazy_polars(x):
    return Table(x).materialize()


@materialize(input_type=pd.DataFrame, version="1.0")
def noop_subtask(x):
    # constant or collection of constants not supported in imperative version
    assert isinstance(x, pd.DataFrame)

    @materialize(input_type=pd.DataFrame, version="1.0")
    def _noop(x):
        return Table(x).materialize()

    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for dataframe tasks.
    return _noop(x)


@materialize(input_type=sa.Table, lazy=True)
def noop_subtask_lazy(x):
    # constant or collection of constants not supported in imperative version
    assert isinstance(x, sa.sql.expression.Alias)

    @materialize(input_type=sa.Table, lazy=True)
    def _noop(x):
        return Table(x).materialize()

    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for same type.
    return _noop(x)


@materialize(input_type=pd.DataFrame, version="1.0")
def noop_subtask_fail_input_type(x):
    # constant or collection of constants not supported in imperative version
    assert isinstance(x, pd.DataFrame)

    @materialize(input_type=sa.Table, version="1.0")
    def _noop(x):
        return Table(x).materialize()

    # calling task with different input type should fail with good error message
    # because we call subtasks without separate dematerialization
    return _noop(x)


@materialize(nout=2, version="1.0", input_type=pd.DataFrame)
def create_tuple(x, y):
    return Table(x).materialize(), Table(y).materialize()


@materialize(input_type=pd.DataFrame)
def assert_table_equal(x, y, **kwargs):
    # This function explicitly has no version set to prevent it from getting cached
    pd.testing.assert_frame_equal(x, y, **kwargs)


@materialize(version="1.0")
def simple_dataframe():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    return Table(df).materialize()


@materialize(version="1.0")
def simple_dataframe_subtask():
    # This is somewhat crazy behavior to call a task within a task. However, since we
    # support calling tasks outside of `with Flow`, this will work for dataframe tasks.
    return simple_dataframe()


@materialize(version="1.0")
def simple_dataframe_debug_materialize_no_taint():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    res = Table(df, name="test_table").materialize()
    return res


@materialize(version="1.0")
def simple_dataframe_debug_materialize_twice():
    df = pd.DataFrame(
        {
            "col1": [0, 1, 2, 3],
            "col2": ["0", "1", "2", "3"],
        }
    )
    tbl = Table(df, name="simple_dataframe_dmt_%%")
    _ = tbl.materialize()
    tbl.obj.iloc[3] = [4, "4"]
    res = tbl.materialize()

    return res


def _get_df_query():
    try:
        unions = [sa.select([sa.literal(i).label("col1"), sa.literal(str(i)).label("col2")]) for i in range(4)]
    except sa.exc.ArgumentError:
        # this works from sqlalchemy 2.0.0 on
        unions = [sa.select(sa.literal(i).label("col1"), sa.literal(str(i)).label("col2")) for i in range(4)]
    return unions[0].union_all(*unions[1:])


@materialize(lazy=True)
def simple_lazy_table():
    query = _get_df_query()
    return Table(query).materialize()


@materialize(lazy=True)
def simple_lazy_table_with_pk():
    query = _get_df_query()
    return Table(query, primary_key="col1").materialize()


@materialize(version="1.0")
def pd_dataframe(data: dict[str, list]):
    return Table(pd.DataFrame(data)).materialize()


@materialize()
def exception(x, r: bool):
    if r:
        raise Exception("THIS EXCEPTION IS EXPECTED")
    return x



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .core import input_stage_versions, materialize

__all__ = [
    "materialize",
    "input_stage_versions",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.pipedag import Flow, Stage, Task, materialize
from pydiverse.pipedag.context import FinalTaskState
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.errors import CacheError
from tests.fixtures.instances import with_instances
from tests.util import select_as
from tests.util import tasks_library as m
from tests.util.spy import spy_task


@with_instances("postgres")
def test_run_specific_task(mocker):
    cache_function_call_allowed = True

    def assert_task_state(
        task_states: dict[Task, FinalTaskState],
        task_name: str,
        expected_state: FinalTaskState,
    ):
        tasks = [t for t in task_states if t.name == task_name]
        assert len(tasks) > 0
        assert all(task_states[t] == expected_state for t in tasks)

    def cache():
        assert cache_function_call_allowed, "Cache function call not allowed with disable_cache_function=True"
        return 0

    @materialize(lazy=True, cache=cache)
    def task_with_cache_fun():
        return select_as(0, "x")

    # We need to assign unique names to these stages, because we can't reuse the
    # same stage lock context between different runs.
    with Flow() as f:
        with Stage("subflow_t1") as s1:
            x1 = m.one()
            x2 = m.two()
            l1 = m.one_sql_lazy()
            task_with_cache_fun()

        with Stage("subflow_t2") as s2:
            y1 = m.create_tuple(x1, x2)
            y2 = m.noop(y1)

    f.run()

    x1_spy = spy_task(mocker, x1)
    x2_spy = spy_task(mocker, x2)
    l1_spy = spy_task(mocker, l1)
    s1_spy = spy_task(mocker, s1.commit_task)
    y1_spy = spy_task(mocker, y1)
    y2_spy = spy_task(mocker, y2)
    s2_spy = spy_task(mocker, s2.commit_task)

    # Run single task separately

    res = f.run(x1)
    x1_spy.assert_called_once()
    x2_spy.assert_not_called()
    l1_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()
    y2_spy.assert_not_called()
    s2_spy.assert_not_called()
    assert_task_state(res.task_states, x1.name, FinalTaskState.COMPLETED)

    res = f.run(y1)
    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    l1_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_called_once()
    y2_spy.assert_not_called()
    s2_spy.assert_not_called()
    assert_task_state(res.task_states, y1.name, FinalTaskState.COMPLETED)

    # Run multiple tasks at once

    res = f.run(x2, y2)
    x1_spy.assert_not_called()
    x2_spy.assert_called_once()
    l1_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()
    y2_spy.assert_called_once()
    s2_spy.assert_not_called()
    assert_task_state(res.task_states, x2.name, FinalTaskState.COMPLETED)
    assert_task_state(res.task_states, y2.name, FinalTaskState.COMPLETED)

    res = f.run(l1)
    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    l1_spy.assert_called_once()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()
    y2_spy.assert_not_called()
    s2_spy.assert_not_called()
    assert_task_state(res.task_states, l1.name, FinalTaskState.COMPLETED)

    cache_function_call_allowed = False

    res = f.run(
        cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
        disable_cache_function=True,
    )
    x1_spy.assert_called_once()
    x2_spy.assert_called_once()
    l1_spy.assert_called_once()
    s1_spy.assert_called_once()
    y1_spy.assert_called_once()
    y2_spy.assert_called_once()
    s2_spy.assert_called_once()
    assert_task_state(res.task_states, x1.name, FinalTaskState.COMPLETED)
    assert_task_state(res.task_states, x2.name, FinalTaskState.COMPLETED)
    assert_task_state(res.task_states, l1.name, FinalTaskState.COMPLETED)
    assert_task_state(res.task_states, y1.name, FinalTaskState.COMPLETED)
    assert_task_state(res.task_states, y2.name, FinalTaskState.COMPLETED)


@with_instances("postgres")
def test_run_specific_task_ambiguous_input(mocker):
    with Flow() as f:
        with Stage("subflow_t1") as s1:
            # The same task appears multiple times in the schema. As a result, we must
            # use the tasks position hash to identify which instance of the noop task
            # we want to use as input for y1
            x1 = m.noop(1)
            x2 = m.noop(2)

        with Stage("subflow_t2") as s2:
            y1 = m.noop(x1)

    f.run()

    x1_spy = spy_task(mocker, x1)
    x2_spy = spy_task(mocker, x2)
    s1_spy = spy_task(mocker, s1.commit_task)
    y1_spy = spy_task(mocker, y1)
    s2_spy = spy_task(mocker, s2.commit_task)

    f.run(y1)

    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_called_once()
    s2_spy.assert_not_called()


@with_instances("postgres")
def test_run_specific_stage(mocker):
    with Flow() as f:
        with Stage("subflow_s1") as s1:
            with Stage("subflow_s2") as s2:
                x1 = m.one()
                x2 = m.two()

            y1 = m.create_tuple(x1, x2)
            y2 = m.noop(y1)

        with Stage("subflow_s3") as s3:
            z1 = m.noop(x1)
            z2 = m.create_tuple(z1, y2)

    f.run()

    x1_spy = spy_task(mocker, x1)
    x2_spy = spy_task(mocker, x2)
    s1_spy = spy_task(mocker, s1.commit_task)
    y1_spy = spy_task(mocker, y1)
    y2_spy = spy_task(mocker, y2)
    s2_spy = spy_task(mocker, s2.commit_task)
    z1_spy = spy_task(mocker, z1)
    z2_spy = spy_task(mocker, z2)
    s3_spy = spy_task(mocker, s3.commit_task)

    # Run single stage

    f.run(s2)  # Inner stage

    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()  # Should be cached
    y2_spy.assert_not_called()  # Should be cached
    s2_spy.assert_called_once()
    z1_spy.assert_not_called()
    z2_spy.assert_not_called()
    s3_spy.assert_not_called()

    f.run(s1)  # Outer stage

    x1_spy.assert_not_called()  # Should be cached
    x2_spy.assert_not_called()  # Should be cached
    s1_spy.assert_called_once()
    y1_spy.assert_not_called()  # Should be cached
    y2_spy.assert_not_called()  # Should be cached
    s2_spy.assert_called_once()
    z1_spy.assert_not_called()
    z2_spy.assert_not_called()
    s3_spy.assert_not_called()

    # Run multiple stags

    f.run(s2, s3)

    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()  # Should be cached
    y2_spy.assert_not_called()  # Should be cached
    s2_spy.assert_called_once()
    z1_spy.assert_not_called()  # Should be cached
    z2_spy.assert_not_called()  # Should be cached
    s3_spy.assert_called_once()


@with_instances("postgres")
def test_run_specific_stage_ambiguous_input(mocker):
    with Flow() as f:
        with Stage("subflow_s1") as s1:
            x1 = m.create_tuple(1, 1)
            x2 = m.create_tuple(x1, x1)

        with Stage("subflow_s2") as s2:
            y1 = m.create_tuple(x1, x2)

    f.run()

    x1_spy = spy_task(mocker, x1)
    x2_spy = spy_task(mocker, x2)
    s1_spy = spy_task(mocker, s1.commit_task)
    y1_spy = spy_task(mocker, y1)
    s2_spy = spy_task(mocker, s2.commit_task)

    # This test is non-trivial, because the task y1 has to fetch its inputs from
    # the cache. However, there are two `create_tuple` tasks in stage s1.
    # Consequently, it isn't clear which of the two tasks should be used.
    # To solve this, we use the position hash to identify which instance of a specific
    # task we are referring to.

    f.run(s2)

    x1_spy.assert_not_called()
    x2_spy.assert_not_called()
    s1_spy.assert_not_called()
    y1_spy.assert_not_called()  # Should be cached
    s2_spy.assert_called_once()


@with_instances("postgres")
def test_run_specific_task_with_table_blob():
    with Flow() as f:
        with Stage("subflow_t1"):
            x1 = m.noop({"x": [1]})
            x2 = m.noop({"x": [2]})
            x3 = m.as_blob("blob")
            x4 = m.noop(x3)

        with Stage("subflow_t2"):
            y1 = m.pd_dataframe(x1)
            y2 = m.pd_dataframe(x2)
            y3 = m.noop((y1, y2, x3, x4))

        with Stage("subflow_t3"):
            z1 = m.noop(y3)
            z2 = m.noop(z1)

    f.run()
    f.run(x4)
    f.run(y1)
    f.run(y3)
    f.run(x1, y2, z1)
    f.run(z2)


@with_instances("postgres")
def test_run_specific_stage_with_table_blob():
    with Flow() as f:
        with Stage("subflow_t1") as s1:
            x1 = m.noop({"x": [1]})
            x2 = m.noop({"x": [2]})
            x3 = m.as_blob("blob")
            x4 = m.noop(x3)

        with Stage("subflow_t2") as s2:
            y1 = m.pd_dataframe(x1)
            y2 = m.pd_dataframe(x2)
            y3 = m.noop((y1, y2, x3, x4))

        with Stage("subflow_t3") as s3:
            z1 = m.noop(y3)
            m.noop(z1)

    f.run(s1)
    f.run(s2)
    f.run(s3)
    f.run(s1, s3)


@with_instances("postgres")
def test_ignore_position_hashes():
    with Flow() as f:
        with Stage("subflow_s1"):
            x1 = m.create_tuple(1, 1)
            x2 = m.create_tuple(x1, x1)

        with Stage("subflow_s2") as s2:
            _ = m.create_tuple(x1, x2)
    f.run()

    # modify flow
    with Flow() as f:
        with Stage("subflow_s1"):
            x1 = m.create_tuple(1, 1)
            x2 = m.create_tuple(2, 2)

        with Stage("subflow_s2") as s2:
            _ = m.create_tuple(x1, x2)

    # as x2 was modified the subflow evaluation should raise a CacheError
    # because an x2 is not found with the correct position hash
    with pytest.raises(CacheError):
        f.run(s2)
    # ignoring the position hash of x2 should make this runnable
    f.run(s2, ignore_position_hashes=True)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import io
import pickle
from io import BytesIO

import pytest
import structlog
from _pytest.capture import EncodedFile

try:
    import dask
except ImportError:
    dask = None


class A(io.TextIOWrapper):
    def __getstate__(self):
        return "a"

    def __reduce__(self):
        return A, (BytesIO(b"hello"),)

    def __reduce_ex__(self, protocol):
        _ = protocol
        return self.__reduce__()


def test_that_io_wrapper_is_pickleable():
    pickle.dumps(A(BytesIO(b"hello")))


def test_that_encoded_file_is_picklable():
    pickle.dumps(EncodedFile(BytesIO(b"hello"), "utf-8"))


@pytest.mark.skipif(dask is None, reason="requires dask")
def test_dask_structlog_configuration_does_not_prevent_pickling():
    def bind_run():
        structlog_config = structlog.get_config()

        def run(parent_futures, **kwargs):
            _ = parent_futures

            structlog.configure(**structlog_config)

            return 1

        run.__name__ = "hi"
        return dask.delayed(run, pure=False)

    results = [bind_run()(parent_futures=[])]
    kw = {
        "traverse": True,
        "optimize_graph": False,
        "scheduler": "processes",
        "num_workers": 8,
        "chunksize": 1,
    }

    dask.compute(results, **kw)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import contextlib

import pytest

from pydiverse.pipedag import ConfigContext


@contextlib.contextmanager
def swallowing_raises(*args, **kwargs):
    with ConfigContext.get().evolve(swallow_exceptions=True):
        with pytest.raises(*args, **kwargs) as raises:
            yield raises



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import types
import warnings

import pandas as pd

from pydiverse.pipedag import Schema, Stage, Table
from pydiverse.pipedag.context import RunContext
from pydiverse.pipedag.errors import CacheError, StageError
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask
from pydiverse.pipedag.materialize.metadata import LazyTableMetadata, TaskMetadata
from pydiverse.pipedag.materialize.store import BaseTableStore
from pydiverse.pipedag.materialize.table_hook_base import CanMatResult, CanRetResult, TableHook


class DictTableStore(BaseTableStore):
    """
    A very basic table store that stores objects in a dictionary.
    Should only ever be used for testing.
    """

    def __init__(self):
        super().__init__()
        self.store = dict()

        self.metadata = dict()
        self.t_metadata = dict()

        self.lazy_table_metadata = dict()
        self.t_lazy_table_metadata = dict()

    def init_stage(self, stage: Stage):
        self.store.setdefault(stage.name, {})
        self.store[stage.transaction_name] = {}

        self.metadata.setdefault(stage, {})
        self.t_metadata[stage] = {}

        self.lazy_table_metadata.setdefault(stage, {})
        self.t_lazy_table_metadata[stage] = {}

    def commit_stage(self, stage: Stage):
        self.store[stage.name] = self.store[stage.transaction_name]
        del self.store[stage.transaction_name]
        self.metadata[stage] = self.t_metadata[stage]
        self.lazy_table_metadata[stage] = self.t_lazy_table_metadata[stage]

    def copy_table_to_transaction(self, table: Table):
        stage = table.stage
        if stage.did_commit:
            raise StageError(
                f"Can't copy table '{table.name}' to transaction. Stage '{stage.name}' has already been committed."
            )

        try:
            t = self.store[stage.name][table.name]
            self.store[stage.transaction_name][table.name] = t
        except KeyError:
            raise CacheError(f"No table with name '{table.name}' found in '{stage.name}' stage") from None

    def copy_lazy_table_to_transaction(self, metadata: LazyTableMetadata, table: Table):
        stage = table.stage
        if stage.did_commit:
            raise StageError(
                f"Can't copy table '{metadata.name}' to transaction. Stage '{stage.name}' has already been committed."
            )

        try:
            RunContext.get().trace_hook.cache_pre_transfer(table)
            t = self.store[stage.name][metadata.name]
            self.store[stage.transaction_name][metadata.name] = t
            RunContext.get().trace_hook.cache_post_transfer(table)
        except KeyError:
            raise CacheError(f"No table with name '{metadata.name}' found in '{stage.name}' stage") from None

    def delete_table_from_transaction(self, table: Table, *, schema: Schema | None = None):
        if schema is None:
            schema = self.get_schema(table.stage.transaction_name)
        try:
            self.store[schema.name].pop(table.name)
        except KeyError:
            return

    def store_task_metadata(self, metadata: TaskMetadata, stage: Stage):
        cache_key = metadata.input_hash + str(metadata.cache_fn_hash)
        self.t_metadata[stage][cache_key] = metadata

    def retrieve_task_metadata(self, task: MaterializingTask, input_hash: str, cache_fn_hash: str) -> TaskMetadata:
        cache_key = input_hash + str(cache_fn_hash)
        try:
            return self.metadata[task.stage][cache_key]
        except KeyError:
            raise CacheError(
                f"There is no metadata for task '{task.name}' with cache key '{task.input_hash}', yet"
            ) from None

    def retrieve_all_task_metadata(
        self, task: MaterializingTask, ignore_position_hashes: bool = False
    ) -> list[TaskMetadata]:
        task_metadata = []
        for m in (
            *self.metadata[task.stage].values(),
            *self.t_metadata[task.stage].values(),
        ):
            if m.name == task.name and ((m.position_hash == task.position_hash) or ignore_position_hashes):
                task_metadata.append(m)
        return task_metadata

    def store_lazy_table_metadata(self, metadata: LazyTableMetadata):
        cache_key = metadata.query_hash + metadata.task_hash
        self.t_lazy_table_metadata[metadata.stage][cache_key] = metadata

    def retrieve_lazy_table_metadata(self, query_hash: str, task_hash: str, stage: Stage) -> LazyTableMetadata:
        try:
            cache_key = query_hash + task_hash
            return self.lazy_table_metadata[stage.name][cache_key]
        except (TypeError, KeyError):
            raise CacheError("Couldn't find metadata for lazy table") from None

    def get_objects_in_stage(self, stage):
        return list(self.store[stage.transaction_name].keys())

    def get_table_objects_in_stage(self, stage: Stage, include_views=True) -> list[str]:
        _ = include_views  # not supported for dict table store
        return list(self.store[stage.transaction_name].keys())


@DictTableStore.register_table(pd)
class PandasTableHook(TableHook[DictTableStore]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        type_ = type(tbl.obj)
        return CanMatResult.new(issubclass(type_, pd.DataFrame))

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        return CanRetResult.new(type_ == pd.DataFrame)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pd.DataFrame],
        stage_name,
        without_config_context: bool = False,
    ):
        if table.name is not None:
            table.obj.attrs["name"] = table.name
        store.store[stage_name][table.name] = table.obj

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type, limit: int | None = None):
        df = store.store[stage_name][table.name]
        if limit:
            df = df[:limit]
        return df.copy()

    @classmethod
    def auto_table(cls, obj: pd.DataFrame):
        if name := obj.attrs.get("name"):
            return Table(obj, name)
        return super().auto_table(obj)


try:
    import pydiverse.transform as pdt
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    pdt = types.ModuleType("pydiverse.transform")
    pdt.Table = None


@DictTableStore.register_table(pdt)
class PydiverseTransformTableHook(TableHook[DictTableStore]):
    @classmethod
    def can_materialize(cls, tbl: Table) -> CanMatResult:
        import pydiverse.transform as pdt
        from pydiverse.transform.extended import build_query

        if not isinstance(tbl, pdt.Table):
            # Not a pydiverse transform table
            return CanMatResult.NO

        query = tbl.obj >> build_query()
        if query is None:
            # SQL is not supported
            return CanMatResult.NO
        else:
            # this error is expected for eager backend
            type_ = type(tbl.obj)
            return CanMatResult.YES_BUT_DONT_CACHE if issubclass(type_, pdt.Table) else CanMatResult.NO

    @classmethod
    def can_retrieve(cls, type_) -> CanRetResult:
        from pydiverse.transform import Pandas, Polars

        return CanRetResult.new(type_ is Polars or type_ is Pandas)

    @classmethod
    def materialize(
        cls,
        store,
        table: Table[pdt.Table],
        stage_name,
        without_config_context: bool = False,
    ):
        from pydiverse.transform.extended import Pandas, collect

        table.obj = table.obj >> collect(Pandas)
        # noinspection PyTypeChecker
        return PandasTableHook.materialize(store, table, stage_name)

    @classmethod
    def retrieve(cls, store, table, stage_name, as_type, limit: int | None = None):
        df = PandasTableHook.retrieve(store, table, stage_name, pd.DataFrame, limit)
        from pydiverse.transform import Polars

        if as_type is Polars:
            import polars as pl

            return pdt.Table(pl.from_pandas(df), name=table.name)
        else:
            return pdt.Table(df, name=table.name)

    @classmethod
    def auto_table(cls, obj: pdt.Table):
        # noinspection PyProtectedMember
        return Table(obj, obj._impl.name)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

from pydiverse.pipedag import Flow, Stage

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, with_instances
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2

pytestmark = [with_instances(DATABASE_INSTANCES)]


@pytest.mark.parametrize(
    "task",
    [
        m.simple_dataframe,
        m.simple_dataframe_with_pk,
        m.simple_dataframe_with_pk2,
        m.simple_dataframe_with_index,
        m.simple_dataframe_with_indexes,
        m.simple_dataframes_with_indexes,
        m.simple_lazy_table,
        m.simple_lazy_table_with_pk,
        m.simple_lazy_table_with_pk2,
        m.simple_lazy_table_with_index,
        m.simple_lazy_table_with_indexes,
        m2.simple_lazy_table,
        m2.simple_lazy_table_with_pk,
    ],
)
def test_materialize_table_with_indexes(task):
    with Flow("flow") as f:
        with Stage("stage"):
            x = task()

            m.assert_table_equal(x, x)
            m.check_pk_length(x)

    assert f.run().successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

try:
    from sqlalchemy import Engine
except ImportError:
    # For compatibility with sqlalchemy < 2.0
    from sqlalchemy.engine import Engine


class PipedagDB2Reflection:
    @staticmethod
    def get_alias_names(engine: Engine, schema: str) -> list[str]:
        """Returns all aliases in a schema"""
        return PipedagDB2Reflection._get_tabname(engine, schema, "A")

    @staticmethod
    def get_nickname_names(engine: Engine, schema: str) -> list[str]:
        """Returns all nicknames in a schema"""
        return PipedagDB2Reflection._get_tabname(engine, schema, "N")

    @staticmethod
    def _get_tabname(engine: Engine, schema: str, _type: str):
        schema = engine.dialect.denormalize_name(schema)
        query = f"""
        SELECT TABNAME
        FROM SYSCAT.TABLES
        WHERE TABSCHEMA = '{schema}' AND TYPE = '{_type}'
        """
        with engine.connect() as conn:
            aliases = conn.exec_driver_sql(query).scalars().all()
        aliases = [engine.dialect.normalize_name(name) for name in aliases]
        return list(aliases)

    @staticmethod
    def get_all_objects(engine: Engine, schema: str):
        schema = engine.dialect.denormalize_name(schema)
        query = f"""
        SELECT TABNAME, TYPE
        FROM SYSCAT.TABLES
        WHERE TABSCHEMA = '{schema}'
        """
        with engine.connect() as conn:
            result = conn.exec_driver_sql(query).all()
        return {name: type_.strip() for name, type_ in result}

    @staticmethod
    def resolve_alias(engine: Engine, name: str, schema: str) -> tuple[str, str]:
        """Recursively resolves an alias

        :returns: A tuple (table_name, schema)
        """

        _schema = engine.dialect.denormalize_name(schema)
        _name = engine.dialect.denormalize_name(name)

        # Recursive CTE query to resolve alias
        query = f"""
        WITH aliases (TABSCHEMA, TABNAME, BASE_TABSCHEMA, BASE_TABNAME, LEVEL) as
            (SELECT a.TABSCHEMA, a.TABNAME, a.BASE_TABSCHEMA, a.BASE_TABNAME, 1
                    FROM SYSCAT.TABLES a
                    WHERE TABSCHEMA = '{_schema}'
                      AND TABNAME = '{_name}'
                      AND TYPE = 'A'
             UNION ALL
             SELECT b.TABSCHEMA, b.TABNAME, b.BASE_TABSCHEMA, b.BASE_TABNAME, r.LEVEL+1
                    FROM aliases r, SYSCAT.TABLES b
                    WHERE r.BASE_TABSCHEMA = b.TABSCHEMA
                      AND r.BASE_TABNAME = b.TABNAME
                      AND b.TYPE = 'A'
                      AND r.LEVEL < 100)
        SELECT BASE_TABNAME, BASE_TABSCHEMA FROM aliases
        ORDER BY LEVEL DESC
        LIMIT 1
        """

        with engine.connect() as conn:
            if result := conn.exec_driver_sql(query).one_or_none():
                return (
                    engine.dialect.normalize_name(result[0]),
                    engine.dialect.normalize_name(result[1]),
                )
        return name, schema


class PipedagMSSqlReflection:
    @staticmethod
    def get_alias_names(engine: Engine, schema: str):
        query = f"""
        SELECT syn.name
        FROM sys.synonyms AS syn
        LEFT JOIN sys.schemas AS schem
               ON syn.schema_id = schem.schema_id
        WHERE schem.name = '{schema}'
        """

        with engine.connect() as conn:
            result = conn.exec_driver_sql(query).scalars().all()
        return result

    @staticmethod
    def resolve_alias(engine: Engine, name: str, schema: str) -> tuple[str, str] | tuple[None, None]:
        from sqlalchemy.dialects.mssql.base import _schema_elements

        query = f"""
        SELECT syn.base_object_name
        FROM sys.synonyms AS syn
        LEFT JOIN sys.schemas AS schem
               ON syn.schema_id = schem.schema_id
        WHERE schem.name = '{schema}'
          AND syn.name = '{name}'
          AND syn.type = 'SN'
        """

        with engine.connect() as conn:
            base_object_name = conn.exec_driver_sql(query).scalar_one_or_none()

        if base_object_name:
            owner, table = _schema_elements(base_object_name)
            return table, owner

        return name, schema

    @staticmethod
    def get_procedure_names(engine: Engine, schema: str):
        query = f"""
        SELECT obj.name
        FROM sys.objects AS obj
        LEFT JOIN sys.schemas AS schem
               ON obj.schema_id = schem.schema_id
        WHERE schem.name = '{schema}'
          AND obj.type = 'P'
        """

        with engine.connect() as conn:
            result = conn.exec_driver_sql(query).scalars().all()
        return result

    @staticmethod
    def get_function_names(engine: Engine, schema: str):
        query = f"""
        SELECT obj.name
        FROM sys.objects AS obj
        LEFT JOIN sys.schemas AS schem
               ON obj.schema_id = schem.schema_id
        WHERE schem.name = '{schema}'
          AND obj.type = 'FN'
        """

        with engine.connect() as conn:
            result = conn.exec_driver_sql(query).scalars().all()
        return result



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
from typing import Any

import pytest
from pandas.core.dtypes.base import ExtensionDtype

from pydiverse.pipedag import (
    AUTO_VERSION,
    ConfigContext,
    Flow,
    Stage,
    Table,
    materialize,
)
from pydiverse.pipedag.backend.table.sql.hooks import PolarsTableHook
from pydiverse.pipedag.backend.table.sql.sql import (
    DISABLE_DIALECT_REGISTRATION,
    SQLTableStore,
)

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import DATABASE_INSTANCES, skip_instances, with_instances
from tests.util.spy import spy_task
from tests.util.sql import get_config_with_table_store
from tests.util.tasks_library import assert_table_equal

pytestmark = [
    pytest.mark.polars,
    with_instances(DATABASE_INSTANCES),
]

try:
    import polars as pl
except ImportError:
    pl = None


def test_table_store():
    @materialize()
    def in_table():
        return Table(
            pl.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                    "str_col": ["a", "abc", "".join("x" for _ in range(1024)), None],
                }
            )
        )

    @materialize()
    def expected_out_table():
        expected_length = 1024
        store = ConfigContext.get().store.table_store
        if store.engine.dialect.name == "mssql":
            expected_length = 256
        return Table(
            pl.DataFrame(
                {
                    "col": [0, 1, 2, 3],
                    "str_col": [
                        "a",
                        "abc",
                        "".join("x" for _ in range(expected_length)),
                        None,
                    ],
                    "x": [1, 1, 1, 1],
                    "y": [2, 2, 2, 2],
                }
            )
        )

    @materialize(input_type=pl.DataFrame)
    def noop(x):
        return Table(x)

    @materialize(lazy=True, input_type=pl.DataFrame)
    def noop_lazy(x):
        return Table(x)

    @materialize(input_type=pl.DataFrame)
    def add_column(x: pl.DataFrame):
        return Table(x.with_columns(pl.lit(1).alias("x")))

    @materialize(lazy=True, input_type=pl.DataFrame)
    def add_column_lazy(x: pl.DataFrame):
        return Table(x.with_columns(pl.lit(2).alias("y")))

    with Flow() as f:
        with Stage("polars"):
            table = in_table()
            table = noop(table)
            table = noop_lazy(table)
            table = add_column(table)
            table = add_column_lazy(table)

            expected = expected_out_table()
            _ = assert_table_equal(table, expected, check_dtype=False)

    assert f.run().successful


def test_auto_version_1(mocker):
    should_swap = True
    value_to_add = 1

    @materialize(version="1.0", nout=2)
    def in_tables():
        in_table_1 = pl.DataFrame({"col": [1, 2, 3, 4]})
        in_table_2 = pl.DataFrame({"col": [4, 3, 2, 1]})

        return Table(in_table_1), Table(in_table_2)

    @materialize(input_type=pl.LazyFrame, version=AUTO_VERSION, nout=2)
    def swap_tables(tbl1, tbl2):
        if should_swap:
            return Table(tbl2), Table(tbl1)
        return Table(tbl1), Table(tbl2)

    @materialize(input_type=pl.LazyFrame, version=AUTO_VERSION)
    def add_value(tbl: pl.LazyFrame):
        return Table(tbl.with_columns(pl.col("col") + value_to_add))

    with Flow() as f:
        with Stage("lazy_polars"):
            x_in, y_in = in_tables()
            x_swap, y_swap = swap_tables(x_in, y_in)

            x_add = add_value(x_swap)
            y_add = add_value(y_swap)

    f.run()

    in_spy = spy_task(mocker, x_in)
    swap_spy = spy_task(mocker, x_swap)
    x_add_spy = spy_task(mocker, x_add)
    y_add_spy = spy_task(mocker, y_add)

    f.run()

    in_spy.assert_not_called()
    # Called once for AUTO_VERSION
    swap_spy.assert_called_once()
    x_add_spy.assert_called_once()
    y_add_spy.assert_called_once()

    should_swap = False
    f.run()

    in_spy.assert_not_called()
    # Called once for AUTO_VERSION, and once again because auto version changed
    swap_spy.assert_called(2)
    x_add_spy.assert_called(2)
    y_add_spy.assert_called(2)

    value_to_add = 2
    f.run()

    in_spy.assert_not_called()
    swap_spy.assert_called_once()
    x_add_spy.assert_called(2)
    y_add_spy.assert_called(2)


def test_auto_version_2(mocker):
    should_swap_inputs = False

    @materialize(input_type=pl.LazyFrame, version=AUTO_VERSION, nout=2)
    def in_tables():
        in_table_1 = pl.LazyFrame({"col": [1, 2, 3, 4]})
        in_table_2 = pl.LazyFrame({"col": [4, 3, 2, 1]})

        if should_swap_inputs:
            return Table(in_table_2), Table(in_table_1)
        else:
            return Table(in_table_1), Table(in_table_2)

    with Flow() as f:
        with Stage("lazy_polars"):
            in_tables_ = in_tables()

    f.run()
    in_tables_spy = spy_task(mocker, in_tables_)

    f.run()
    in_tables_spy.assert_called_once()

    should_swap_inputs = True
    f.run()
    in_tables_spy.assert_called(2)


@skip_instances("parquet_backend")
def test_custom_download():
    class TestTableStore(ConfigContext.get().store.table_store.__class__):
        _dialect_name = DISABLE_DIALECT_REGISTRATION
        # this subclass is just to make sure hooks of other tests are not affected
        pass

    cfg = get_config_with_table_store(ConfigContext.get(), TestTableStore)

    import numpy as np

    @TestTableStore.register_table(pl, replace_hooks=[PolarsTableHook])
    class CustomPolarsDownloadTableHook(PolarsTableHook):
        @classmethod
        def download_table(
            cls,
            query: Any,
            store: SQLTableStore,
            dtypes: dict[str, ExtensionDtype | np.dtype] | None = None,
        ) -> pl.DataFrame:
            # to simplify this test with various dependencies and platforms, it is
            # easier to use pandas:
            import pandas as pd

            connection_uri = store.engine_url.render_as_string(hide_password=False)
            pandas_df = pd.read_sql(query, con=connection_uri)
            # # pl.read_database_uri fails for duckdb and osx-arm64
            # # (newer conda-forge builds for connectorx for osx-arm64 are broken)
            # df = pl.read_database_uri(query, connection_uri)
            df = pl.from_pandas(pandas_df)
            return df.with_columns(pl.lit(True).alias("custom_download"))

    @materialize()
    def numpy_input():
        df = pl.DataFrame(
            {
                "col": [0, 1, 2, 3],
            }
        )
        return Table(df)

    @materialize(input_type=pl.DataFrame)
    def verify_custom(t):
        assert "custom_download" in t.columns

    with Flow() as f:
        with Stage("stage"):
            t = numpy_input()
            verify_custom(t)

    assert f.run(config=cfg).successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import structlog

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    boto3 = None
    ClientError = None


def create_bucket_if_not_exists(bucket_name: str, s3_client, region: str | None = None):
    logger = structlog.get_logger(__name__ + ":create_bucket_if_not_exists")
    try:
        s3_client.head_bucket(Bucket=bucket_name)
        logger.info(f"Bucket '{bucket_name}' already exists.")
    except ClientError as e:
        error_code = int(e.response["Error"]["Code"])
        if error_code == 404:
            # Bucket does not exist, safe to create
            s3_client.create_bucket(
                Bucket=bucket_name,
                CreateBucketConfiguration={"LocationConstraint": region} if region is not None else {},
            )
            logger.info(f"Bucket '{bucket_name}' created.")
        elif error_code == 403:
            logger.info(f"Bucket '{bucket_name}' exists but is owned by someone else.")
        else:
            raise


def initialize_test_s3_bucket():
    if boto3 is not None:
        minio_client = boto3.client(
            "s3",
            endpoint_url="http://localhost:9000",
            aws_access_key_id="minioadmin",  # default MinIO credentials
            aws_secret_access_key="minioadmin",
            region_name="us-east-1",
            config=boto3.session.Config(s3={"addressing_style": "path"}),
        )
        create_bucket_if_not_exists("pipedag-test-bucket", minio_client)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd

from pydiverse.pipedag import Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode
from tests.fixtures.instances import with_instances

dfA_source = pd.DataFrame(
    {
        "a": [0, 1, 2, 4],
        "b": [9, 8, 7, 6],
    }
)
dfA = dfA_source.copy()
input_hash = hash(str(dfA))


def has_new_input(dummy_arg):
    """Returns whether new input is available via input hash.

    :param dummy_arg: Argument used to test that custom cache invalidation function
        gets same arguments as task function
    :return: hash value of input (stored hash must not exactly be input hash)
    """
    assert dummy_arg == "irrelevant"
    global input_hash
    return input_hash


# noinspection DuplicatedCode
@materialize(nout=2, cache=has_new_input, version="1.0")
def input_task(dummy_arg):
    global dfA
    return Table(dfA, "dfA"), Table(dfA, "dfB")


@materialize(input_type=pd.DataFrame, version="1.0")
def double_values(df: pd.DataFrame):
    return Table(df.transform(lambda x: x * 2))


# noinspection PyTypeChecker
def get_flow():
    with Flow("FLOW") as flow:
        with Stage("stage_1"):
            dummy_arg = "irrelevant"
            a, b = input_task(dummy_arg)
            a2 = double_values(a)

        with Stage("stage_2"):
            b2 = double_values(b)
            a3 = double_values(a2)

    return flow, b2, a3


@with_instances("postgres")
def test_source_invalidation():
    # trigger reload of input data
    global dfA
    global input_hash

    flow, out1, out2 = get_flow()

    with StageLockContext():
        result = flow.run()
        assert result.successful

        v_out1, v_out2 = result.get(out1), result.get(out2)
        pd.testing.assert_frame_equal(dfA_source * 2, v_out1, check_dtype=False)
        pd.testing.assert_frame_equal(dfA_source * 4, v_out2, check_dtype=False)

    # modify input without updating input hash => cached version is used
    dfA["a"] = 10 + dfA_source["a"]

    # this run should work from caches and not change outputs
    with StageLockContext():
        result = flow.run()
        assert result.successful

        v_out1, v_out2 = result.get(out1), result.get(out2)
        pd.testing.assert_frame_equal(dfA_source * 2, v_out1, check_dtype=False)
        pd.testing.assert_frame_equal(dfA_source * 4, v_out2, check_dtype=False)

    # update input hash trigger reload of new input data
    input_hash = hash(str(dfA))

    with StageLockContext():
        # this run should ignore fresh input at source nodes and not change outputs
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.successful

        v_out1, v_out2 = result.get(out1), result.get(out2)
        pd.testing.assert_frame_equal(dfA_source * 2, v_out1, check_dtype=False)
        pd.testing.assert_frame_equal(dfA_source * 4, v_out2, check_dtype=False)

    with StageLockContext():
        result = flow.run()
        assert result.successful

        v_out1, v_out2 = result.get(out1), result.get(out2)

        pd.testing.assert_frame_equal(dfA * 2, v_out1, check_dtype=False)
        pd.testing.assert_frame_equal(dfA * 4, v_out2, check_dtype=False)


if __name__ == "__main__":
    test_source_invalidation()
